﻿using System;
using System.Collections.Generic;
using net.citrusleaf;

namespace Citrusleaf.Demo
{
    class Add : Example
    {
        public Add(Console console)
            : base(console)
        {
        }

        /// <summary>
        /// Add integer values.
        /// </summary>
        public override void RunExample(CitrusleafClient cs, string host, int port, string ns, string set, bool singleBin)
        {
            InitializeAndAdd(cs, ns, set, singleBin);
            ContinuousAdd(cs, ns, set, singleBin);
        }

        /// <summary>
        /// Delete and add integers.
        /// </summary>
        private void InitializeAndAdd(CitrusleafClient cs, string ns, string set, bool singleBin)
        {
            string key = "addkey";
            string bin = singleBin ? "" : "addbin";  // Single bin servers don't need a bin name.
            int value = 10;

            // Delete record if it already exists.
            CitrusleafResult r = cs.Delete(ns, set, key, null);

            if (!(r == CitrusleafResult.OK || r == CitrusleafResult.KEY_NOT_FOUND_ERROR))
                throw new Exception(string.Format("Failed to delete: ns={0} set={1} key={2} code={3}",
                    ns, set, key, r));

            console.Info("Initial add will create record.  Initial value is " + value + '.');
            r = cs.Add(ns, set, key, bin, value, null);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to add: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                    ns, set, key, bin, value, r));

            value = 5;
            console.Info("Add " + value + " to existing record.");
            r = cs.Add(ns, set, key, bin, value, null);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to add: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                    ns, set, key, bin, value, r));

            int expected = 15;
            object received;
            uint generation;
            r = cs.Get(ns, set, key, bin, out received, out generation, DEFAULT_TIMEOUT_MS);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to get: ns={0} set={1} key={2} bin={3} code={4}",
                    ns, set, key, bin, r));

            // The value received from the server is an unsigned byte stream.
            // Convert to an integer before comparing with expected.
            int rec = (int)(uint)received;
            if (rec == expected)
            {
                console.Info("Add successful: ns={0} set={1} key={2} bin={3} value={4}",
                    ns, set, key, bin, received);
            }
            else
            {
                console.Error("Add mismatch: Expected {0}. Received {1}.", expected, received);
            }
        }

        /// <summary>
        /// Revolving integer add without deletion.
        /// </summary>
        private void ContinuousAdd(CitrusleafClient cs, string ns, string set, bool singleBin)
        {
            string key = "continuouskey";
            string bin = singleBin ? "" : "continuousbin";  // Single bin servers don't need a bin name.
            int value1 = 0;
            int value2 = 5;

            // get old value if it is there
            object received;
            uint generation;
            CitrusleafResult r = cs.Get(ns, set, key, bin, out received, out generation, 100);
            if (r == CitrusleafResult.OK)
            {
                value1 = (int)(uint)received;
                console.Info("Existing value is " + value1 + '.');
            }
            else if (r == CitrusleafResult.KEY_NOT_FOUND_ERROR)
            {
                console.Info("No existing record found. Add will create record value of  " + value2 + '.');
            }
            else
            {
                throw new Exception(string.Format("Failed to Get: ns={0} set={1} key={2} code={3}",
                    ns, set, key, r));
            }

            console.Info("Add " + value2 + " to record.");
            r = cs.Add(ns, set, key, bin, value2, null);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to add: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                    ns, set, key, bin, value2, r));

            int expected = value1 + value2;
            r = cs.Get(ns, set, key, bin, out received, out generation, DEFAULT_TIMEOUT_MS);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to get: ns={0} set={1} key={2} bin={3} code={4}",
                    ns, set, key, bin, r));

            // The value received from the server is an unsigned byte stream.
            // Convert to an integer before comparing with expected.
            int rec = (int)(uint)received;
            if (rec == expected)
            {
                console.Info("Add successful: ns={0} set={1} key={2} bin={3} value={4}",
                    ns, set, key, bin, received);
            }
            else
            {
                console.Error("Add mismatch: Expected {0}. Received {1}.", expected, received);
            }
        }
    }
}
