﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using net.citrusleaf;
using System.Windows.Forms;

namespace Citrusleaf.Demo
{
    /// <summary>
    /// Console logger with RichTextBox GUI target.
    /// </summary>
    class RichTextBoxConsole : Console
    {
        private RichTextBox box;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="box">Target GUI control.</param>
        public RichTextBoxConsole(RichTextBox box)
        {
            this.box = box;
        }

        protected override void Write(string message)
        {
            if (box.InvokeRequired)
            {
                box.Invoke(new Action(delegate { WriteLog(message); }));
                return;
            }
            else
            {
                WriteLog(message);
            }
        }

        private void WriteLog(string message)
        {
            box.SelectionStart = box.Text.Length;
            box.AppendText(message);
            box.ScrollToCaret();
            box.AppendText(System.Environment.NewLine);
            box.Refresh();
        }

        public override void Clear()
        {
            if (box.InvokeRequired)
            {
                box.Invoke(new Action(delegate { ClearLog(); }));
                return;
            }
            else
            {
                ClearLog();
            }
        }

        private void ClearLog()
        {
            box.Clear();
            box.Refresh();
        }
    }

    /// <summary>
    /// Log base class.  To create a new logging target, inherit from this class and override Write().
    /// </summary>
    abstract class Console
    {
        /// <summary>
        /// Log info message.
        /// </summary>
        public void Info(string format, params object[] args)
        {
            Log(CitrusleafClient.LogLevel.INFO, format, args);
        }

        /// <summary>
        /// Log warning message.
        /// </summary>
        public void Warn(string format, params object[] args)
        {
            Log(CitrusleafClient.LogLevel.WARNING, format, args);
        }

        /// <summary>
        /// Log info message.
        /// </summary>
        public void Error(string format, params object[] args)
        {
            Log(CitrusleafClient.LogLevel.ERROR, format, args);
        }

        /// <summary>
        /// Log message.
        /// </summary>
        public void Log(CitrusleafClient.LogLevel level, string format, params object[] args)
        {
            Write(DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss") + ' ' + level.ToString() + ' ' + string.Format(format, args));
        }

        /// <summary>
        /// Log line without timestamp or level.
        /// </summary>
        public void Log(string format, params object[] args)
        {
            Write(string.Format(format, args));
        }

        protected abstract void Write(string message);
        public abstract void Clear();
    }
}
