﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;
using System.Threading;

namespace Citrusleaf.Demo
{
    public partial class DemoForm : Form
    {
        private Thread thread;
        private volatile ExampleTreeNode currentExample;

        public DemoForm()
        {
            InitializeComponent();
            FormInit();
        }

        private void FormInit()
        {
            try
            {
                hostBox.Text = Properties.Settings.Default.Host;
                portBox.Text = Properties.Settings.Default.Port.ToString();
                nsBox.Text = Properties.Settings.Default.Namespace;
                setBox.Text = Properties.Settings.Default.Set;

                codeBox.Font = new Font("Consolas", 10.0f);

                Console console = new RichTextBoxConsole(consoleBox);
                TreeNode info = new ExampleTreeNode("Server Info", new ServerInfo(console));

                TreeNode examples = new TreeNode("Examples", new TreeNode[] {
                    info,
                    new ExampleTreeNode("Set/Get", new SetGet(console)),
                    new ExampleTreeNode("Add", new Add(console)),
                    new ExampleTreeNode("Batch Get", new BatchGet(console)),
                    new ExampleTreeNode("Generation", new Generation(console)),
                    new ExampleTreeNode("Serialize", new Serialize(console)),
                    new ExampleTreeNode("Touch", new Touch(console))
                });
                TreeNode benchmarks = new TreeNode("Benchmarks", new TreeNode[] {
                    new ExampleTreeNode("Linear Write/Read", new LinearWriteRead(console)),
                    new ExampleTreeNode("Multi-Threaded Load", new MultiThreadedLoad(console)),
                    new ExampleTreeNode("Scan", new Scan(console)),
                    new ExampleTreeNode("ScanSeries", new ScanSeries(console))
                });

                examplesView.Nodes.Add(examples);
                examplesView.Nodes.Add(benchmarks);
                examplesView.SelectedNode = info;
                examplesView.ExpandAll();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void FormClose(object sender, FormClosingEventArgs e)
        {
            try
            {
                Properties.Settings.Default.Host = hostBox.Text.Trim();
                Properties.Settings.Default.Port = int.Parse(portBox.Text);
                Properties.Settings.Default.Namespace = nsBox.Text.Trim();
                Properties.Settings.Default.Set = setBox.Text.Trim();
                Properties.Settings.Default.Save();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void RunExample(object sender, MouseEventArgs e)
        {          
            try
            {
                TreeNode node = examplesView.SelectedNode;

                if (node == null || ! (node is ExampleTreeNode))
                    throw new Exception("Please select an example program.");

                if (currentExample != null)
                    throw new Exception("Stop the current example before running a new example.");

                currentExample = (ExampleTreeNode)node;
                thread = new Thread(RunExampleThread);
                thread.Start();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void RunExampleThread()
        {
            try
            {
                int port = int.Parse(portBox.Text);
                currentExample.Run(hostBox.Text.Trim(), port, nsBox.Text.Trim(), setBox.Text.Trim());
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                currentExample = null;
            }
        }

        private void StopExample(object sender, EventArgs e)
        {
            try
            {
                if (currentExample != null)
                {
                    currentExample.Stop();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void ExampleSelected(object sender, TreeViewEventArgs e)
        {
            try
            {
                TreeNode node = examplesView.SelectedNode;

                if (node == null || !(node is ExampleTreeNode))
                    return;

                codeBox.Clear();
                ExampleTreeNode example = (ExampleTreeNode)node;
                codeBox.Text = example.Read();
                HighlightSourceCode();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// Extremely simple syntax highlighter.
        /// </summary>
        private void HighlightSourceCode()
        {
            string tokens = @"(using|namespace|class|public|base|override|void|string|int|bool|\stry\s|null|if|throw|new|\sobject\s|uint|out|else|catch|finally|private|foreach|\sin\s)";
            HighlightSegments(tokens, Color.Blue);

            tokens = "(\".*\"|'.?')";
            HighlightSegments(tokens, Color.DarkRed);

            tokens = "//.*";
            HighlightSegments(tokens, Color.Green);

            int offset = codeBox.Text.IndexOf("RunExample");
            codeBox.SelectionStart = offset;
            codeBox.ScrollToCaret();
            codeBox.Refresh();
        }

        private void HighlightSegments(string tokens, Color color)
        {
            Regex regex = new Regex(tokens);
            MatchCollection mc = regex.Matches(codeBox.Text);

            foreach (Match m in mc)
            {
                int startIndex = m.Index;
                int stopIndex = m.Length;
                codeBox.Select(startIndex, stopIndex);
                codeBox.SelectionColor = color;
            }
        }
    }

    class ExampleTreeNode : TreeNode
    {
        private Example example;

        public ExampleTreeNode(String text, Example example)
            : base(text)
        {
            this.example = example;
        }

        public string Read()
        {
            string path = @"..\..\" + example.GetType().Name + ".cs";
            return File.ReadAllText(path);
        }

        public void Run(string host, int port, string ns, string set)
        {
            example.Run(host, port, ns, set);
        }

        public void Stop()
        {
            example.Stop();
        }
    }
}
