﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using net.citrusleaf;

namespace Citrusleaf.Demo
{
    class Generation : Example
    {
        public Generation(Console console)
            : base(console)
        {
        }

        /// <summary>
        /// Exercise record generation functionality.
        /// </summary>
        public override void RunExample(CitrusleafClient cs, string host, int port, string ns, string set, bool singleBin)
        {
            string key = "genkey";
            string bin = singleBin ? "" : "genbin";  // Single bin servers don't need a bin name.
            String value = "genvalue1";

            // Delete record if it already exists.
            CitrusleafResult r = cs.Delete(ns, set, key, null);

            if (!(r == CitrusleafResult.OK || r == CitrusleafResult.KEY_NOT_FOUND_ERROR))
                throw new Exception(string.Format("Failed to delete: ns={0} set={1} key={2} code={3}",
                    ns, set, key, r));

            // Set some values for the same record.
            console.Info("Set: ns={0} set={1} key={2} bin={3} value={4}",
                ns, set, key, bin, value);

            r = cs.Set(ns, set, key, bin, value, null);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to set: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                    ns, set, key, bin, value, r));

            value = "genvalue2";
            console.Info("Set: ns={0} set={1} key={2} bin={3} value={4}",
                ns, set, key, bin, value);

            r = cs.Set(ns, set, key, bin, value, null);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to set: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                    ns, set, key, bin, value, r));

            // Retrieve record and its generation count.
            object received;
            uint generation;
            r = cs.Get(ns, set, key, bin, out received, out generation, DEFAULT_TIMEOUT_MS);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to get: ns={0} set={1} key={2} bin={3} code={4}",
                    ns, set, key, bin, r));

            if (received.Equals(value)) 
            {
                console.Info("Get successful: ns={0} set={1} key={2} bin={3} value={4} generation={5}", 
                    ns, set, key, bin, received, generation);
            }
            else 
            {
                throw new Exception(string.Format(
                    "Get mismatch: Expected %s. Received %s.", value, received));
            }

            // Set record and fail if it's not the expected generation.
            value = "genvalue3";
            console.Info("Set: ns={0} set={1} key={2} bin={3} value={4} expected generation={5}",
                ns, set, key, bin, value, generation);

            CitrusleafWriteInfo writeInfo = new CitrusleafWriteInfo(DEFAULT_TIMEOUT_MS);
            writeInfo.Generation = generation;
            r = cs.Set(ns, set, key, bin, value, writeInfo);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to set: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                    ns, set, key, bin, value, r));

            // Set record with invalid generation and check results .
            value = "genvalue4";
            generation = 9999;
            console.Info("Set: namespace={0} set={1} key={2} bin={3} value={4} expected generation={5}",
                ns, set, key, bin, value, generation);

            writeInfo.Generation = generation;
            r = cs.Set(ns, set, key, bin, value, writeInfo);

            if (r == CitrusleafResult.GENERATION_ERROR) 
            {
                console.Info("Success: Generation error returned as expected.");
            }
            else 
            {
                throw new Exception(String.Format(
                    "Unexpected set return code: ns={0} set={1} key={2} bin={3} value={4} code={5}", 
                    ns, set, key, bin, value, r));
            }

            // Verify results.
            r = cs.Get(ns, set, key, bin, out received, out generation, DEFAULT_TIMEOUT_MS);

            if (r != CitrusleafResult.OK)
                throw new Exception(string.Format(
                    "Failed to get: ns={0} set={1} key={2} bin={3} code={4}",
                    ns, set, key, bin, r));

            value = "genvalue3";
            if (received.Equals(value)) 
            {
                console.Info(
                    "Get successful: ns={0} set={1} key={2} bin={3} value={4} generation={5}",
                    ns, set, key, bin, received, generation);
            }
            else 
            {
                throw new Exception(String.Format(
                    "Get mismatch: Expected %s. Received %s.", value, received));
            }
        }
    }
}
