﻿using System;
using System.Collections.Generic;
using net.citrusleaf;

namespace Citrusleaf.Demo
{
    class LinearWriteRead : Example
    {
        public LinearWriteRead(Console console)
            : base(console)
        {
        }

        /// <summary>
        /// Write/Read large blocks of data and measure performance.
        /// </summary>
        public override void RunExample(CitrusleafClient cs, string host, int port, string ns, string set, bool singleBin)
        {
            int size = 100000;
            string keyPrefix = "linearkey";
            string bin = singleBin ? "" : "linearbin";  // Single bin servers don't need a bin name.
            writeRecords(cs, ns, set, keyPrefix, bin, size);
            readRecords(cs, ns, set, keyPrefix, bin, size);
        }

        private void writeRecords
        (
            CitrusleafClient cs, 
            string ns, 
            string set,
            string keyPrefix,
            string bin, 
            int size
        )
        {
            if (!valid)
                return;

            // Write multiple records.
            console.Info("Write " + size + " records.");
            CitrusleafResult r;
            DateTime begin = DateTime.Now;
            int i = 1;

            for (; i <= size && valid; i++)
            {
                string key = keyPrefix + i;
                string value = i.ToString();

                r = cs.Set(ns, set, key, bin, value, null);

                if (r != CitrusleafResult.OK)
                    throw new Exception(string.Format(
                        "Failed to set: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                        ns, set, key, bin, value, r));

                if (i % 10000 == 0)
                {
                    console.Info("Records " + i);
                }
            }
            DateTime end = DateTime.Now;
            TimeSpan ts = end.Subtract(begin);
            size = i - 1;
            console.Info("Records written: " + size);
            console.Info("Elapsed time: " + ts.ToString());
            double performance = Math.Round(((double)size) / ((double)ts.TotalSeconds), 0);
            console.Info("Writes/second: " + performance);
        }

        private void readRecords
        (
            CitrusleafClient cs, 
            string ns, 
            string set,
            string keyPrefix,
            string bin, 
            int size
        )
        {
            if (!valid)
                return;

            // Read multiple records.
            console.Info("Read " + size + " records.");
            CitrusleafResult r;
            object value;
            uint generation;
            DateTime begin = DateTime.Now;
            int i = 1;

            for (; i <= size && valid; i++)
            {
                string key = keyPrefix + i;
                string expected = i.ToString();

                r = cs.Get(ns, set, key, bin, out value, out generation, DEFAULT_TIMEOUT_MS);

                if (r != CitrusleafResult.OK)
                    throw new Exception(string.Format(
                        "Failed to get: ns={0} set={1} key={2} bin={3} code={4}",
                        ns, set, key, bin, r));

                if (! value.Equals(expected))
                {
                    console.Error("Set/Get mismatch: Expected {0}. Received {1}.",
                        expected, value);
                }

                if (i % 10000 == 0)
                {
                    console.Info("Records " + i);
                }
            }
            DateTime end = DateTime.Now;
            TimeSpan ts = end.Subtract(begin);
            size = i - 1;
            console.Info("Records read: " + size);
            console.Info("Elapsed time: " + ts.ToString());
            double performance = Math.Round(((double)size) / ((double)ts.TotalSeconds), 0);
            console.Info("Reads/second: " + performance);
        }
    }
}
