﻿using System;
using System.Collections.Generic;
using System.Threading;
using net.citrusleaf;

namespace Citrusleaf.Demo
{
    class MultiThreadedLoad : Example
    {
        private const int UNSET = -1;
        private const int DELETED = -2;
        private const int INUSE = -3;
        private static CitrusleafWriteInfo WriteInfo = new CitrusleafWriteInfo(2000);

        private string host;
        private int port;
        private string ns;
        private string set;
        private string bin;

        private Thread[] threads;
        private Thread tickerThread;
        private CitrusleafClient cs;
        private Random random;

        private object valuesLock;
        private int[] values;

        // These are counters available across all threads.
        // Use with Interlocked.increment
        private int writeCounter;
        private int deleteCounter;
        private int readCounter;

        public MultiThreadedLoad(Console console)
            : base(console)
        {
        }

        /// <summary>
        /// Write/Read large blocks of data using multiple threads and measure performance.
        /// </summary>
        public override void RunExample(CitrusleafClient cs, string host, int port, string ns, string set, bool singleBin)
        {
            this.cs = cs;
            this.host = host;
            this.port = port;
            this.ns = ns;
            this.set = set;
            this.bin = singleBin ? "" : "demobin";  // Single bin servers don't need a bin name.

            int objectSize = 1000000;
            values = new int[objectSize];

            valuesLock = new Object();
            random = new Random();

            for (int i = 0; i < objectSize; i++)
                values[i] = UNSET;

            writeCounter = 0;
            readCounter = 0;
            deleteCounter = 0;

            StartThreads(30);
        }

        private void StartThreads(int threadMax)
        {
            tickerThread = new Thread(new ThreadStart(this.Ticker));
            tickerThread.Start();

            threads = new Thread[threadMax];
            for (int i = 0; i < threadMax; i++)
            {
                threads[i] = new Thread(new ThreadStart(this.Worker));
                threads[i].Start();
            }

            for (int i = 0; i < threadMax; i++)
            {
                threads[i].Join();
            }
            valid = false;
            tickerThread.Join();
        }

        private void Ticker()
        {
            try
            {
                RunTicker();
            }
            catch (Exception ex)
            {
                console.Error(ex.Message);
            }
        }

        private void RunTicker()
        {
            DateTime prevTime = DateTime.Now;
            int prevTotal = 0;

            while (valid)
            {
                int r = readCounter;
                int w = writeCounter;
                int d = deleteCounter;
                int total = r + w + d;
                DateTime time = DateTime.Now;
                double seconds = time.Subtract(prevTime).TotalSeconds;

                if (seconds > 0.0)
                {
                    double opsPerSecond = Math.Round(((double)(total - prevTotal)) / seconds, 0);

                    console.Info("Read {0} Write {1} Delete {2} Total {3} Operations/Second {4}",
                        r, w, d, total, opsPerSecond);
        
                    prevTotal = total;
                    prevTime = time;
                }
                Thread.Sleep(1000);
            }
        }

        private void Worker()
        {
            try
            {
                RunWorker();
            }
            catch (Exception ex)
            {
                console.Error(ex.Message);
            }
        }

        private void RunWorker()
        {
            while (valid)
            {
                // Choose key at random.
                int value;
                int key;
                int ops = 0;

                lock (valuesLock)
                {
                    do
                    {
                        if (++ops >= values.Length)
                            throw new Exception("All keys in use.");

                        key = random.Next(0, values.Length);
                    } while (values[key] == INUSE);

                    value = values[key];
                    values[key] = INUSE;
                }

                try
                {
                    // Process corresponding value.
                    switch (value)
                    {
                        case UNSET:
                            value = Write(key);
                            break;

                        case DELETED:
                            value = ProcessDeleted(key, value);
                            break;

                        default:
                            value = ProcessNormal(key, value);
                            break;
                    }
                }
                finally
                {
                    lock (valuesLock)
                    {
                        values[key] = value;
                    }
                }
            }
        }

        private int ProcessNormal(int offset, int expected)
        {
            if (Read(offset, expected) == UNSET)
                return UNSET;

            // Roll a die.
            int die;
            lock (valuesLock)
            {
                die = random.Next(0, 3);
            }

            // Delete record 25% of the time.
            if (die == 0)
            {
                return Delete(offset);
            }

            // Update record 50% of the time.
            if (die <= 2)
            {
                return Write(offset);
            }

            // Leave record alone 25% of the time.
            return expected;
        }

        private int ProcessDeleted(int offset, int expected)
        {
            object value;
            uint generation;
            CitrusleafResult r = cs.Get(ns, set, offset.ToString(), bin, out value, out generation, 2000);

            if (r != CitrusleafResult.KEY_NOT_FOUND_ERROR)
            {
                console.Warn("Read deleted record: ns={0} set={1} key={2} bin={3} code={4}",
                    ns, set, offset, bin, r);
            }
            Interlocked.Increment(ref readCounter);

            // Roll a die.
            int die;
            lock (valuesLock)
            {
                die = random.Next(0, 2);
            }

            // Write over deleted value 67% of the time.
            if (die <= 1)
            {
                return Write(offset);
            }

            // Leave record alone 33% of the time.
            return expected;
        }

        private int Write(int key)
        {
            int value;
            lock (valuesLock)
            {
                value = random.Next(0, Int32.MaxValue);
            }

            CitrusleafResult r = cs.Set(ns, set, key.ToString(), bin, value.ToString(), WriteInfo);

            if (r != CitrusleafResult.OK)
            {
                if (r != CitrusleafResult.TIMEOUT)
                {
                    bool stopWrites = GetIsStopWrites(host, port, ns);

                    if (stopWrites)
                    {
                        if (valid)
                        {
                            console.Error("Server is currently in readonly mode. Shutting down...");
                            valid = false;
                        }
                    }
                    else
                    {
                        console.Error("Failed to set: ns={0} set={1} key={2} bin={3} value={4} code={5}",
                            ns, set, key.ToString(), bin, value.ToString(), r);
                    }
                }
                return UNSET;
            }
            Interlocked.Increment(ref writeCounter);
            return value;
        }

        private int Read(int key, int expected)
        {
            object value;
            uint generation;
            CitrusleafResult r = cs.Get(ns, set, key.ToString(), bin, out value, out generation, 2000);

            if (r != CitrusleafResult.OK)
            {
                console.Error("Failed to get: ns={0} set={1} key={2} bin={3} code={4}",
                    ns, set, key.ToString(), bin, r);
                return UNSET;
            }

            if (Convert.ToInt32(value) != expected)
            {
                console.Error("Value mismatch: Expected {0}. Received {1}.",
                    expected, value);
            }
            Interlocked.Increment(ref readCounter);
            return expected;
        }

        private int Delete(int key)
        {
            CitrusleafResult r = cs.Delete(ns, set, key.ToString(), WriteInfo);

            if (r != CitrusleafResult.OK)
            {
                console.Error("Failed to delete: ns={0} set={1} key={2} code={3}",
                    ns, set, key.ToString(), r);
                return UNSET;
            }
            Interlocked.Increment(ref deleteCounter);
            return DELETED;
        }
    }
}
