using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Security.Cryptography;
using System.Runtime.Serialization;
using System.Collections.Concurrent;

namespace net.citrusleaf
{
    public class Buffer
    {
        enum Op
        {
            READ = 1,
            WRITE = 2,
            WRITE_UNIQUE = 3, // write a namespace wide unique value (ints, strings, globs)
            WRITE_NOW = 4,    // write the current time
            ADD = 5,          // add the bin value to whatever is there
            APPEND = 6,  // append the value to the list
            APPEND_EXT = 7,
            APPEND_QUERY = 8,
            TOUCH = 11
        }

        /// <summary>
        /// bitfields are better off as actual integers
        /// </summary>
        const int CITRUSLEAF_INFO1_READ	=			(1 << 0);		// contains a read operation
        const int CITRUSLEAF_INFO1_GET_ALL =		(1 << 1); 		// get all bins, period
        const int CITRUSLEAF_INFO1_GET_ALL_NODATA =	(1 << 2); 		// get all bins WITHOUT data (currently unimplemented)
        const int CITRUSLEAF_INFO1_VERIFY =    		(1 << 3); 		// verify is a GET transaction that includes data, and assert if the data aint right

        const int CITRUSLEAF_INFO2_WRITE =				(1 << 0);		// contains a write semantic
        const int CITRUSLEAF_INFO2_DELETE =			    (1 << 1);  		// delete
        const int CITRUSLEAF_INFO2_GENERATION =			(1 << 2); 		// pay attention to the generation
        const int CITRUSLEAF_INFO2_GENERATION_GT =		(1 << 3); 		// apply write if new generation >= old, good for restore
        const int CITRUSLEAF_INFO2_GENERATION_DUP =  	(1 << 4);		// if a generation collision, create a duplicate
        const int CITRUSLEAF_INFO2_WRITE_UNIQUE =	    (1 << 5); 		// write only if it doesn't exist
        const int CITRUSLEAF_INFO2_WRITE_BINUNIQUE =	(1 << 6);

        const int CITRUSLEAF_INFO3_LAST =     			(1 << 0);     	// this is the last of a multi-part message
        const int CITRUSLEAF_INFO3_TRACE =				(1 << 1);		// apply server trace logging for this transaction
        const int CITRUSLEAF_INFO3_TOMBSTONE =          (1 << 2);		// if set on response, a version was a delete tombstone

        enum ParticleType
        {
            NULL = 0,
            INTEGER = 1,
            BIGNUM = 2,
            STRING = 3,
            BLOB = 4,
            TIMESTAMP = 5,
            DIGEST = 6,
            JAVA_BLOB = 7,
            CSHARP_BLOB = 8,
            PYTHON_BLOB = 9,
            RUBY_BLOB = 10,
            PHP_BLOB = 11,
            ERLANG_BLOB = 12,
            SEGMENT_POINTER_UNUSED = 13, // CL internal type only for RTA
            LIST = 14,
            DICT = 15,
            APPEND_DICT = 16,
            APPEND_LIST = 17,
        }

        [ThreadStatic]
        static HashAlgorithm hash_algorithm;

        private static HashAlgorithm getHashAlgorithm()
        {
            if (hash_algorithm == null)
            {
                hash_algorithm = RIPEMD160Managed.Create();
            }

            return hash_algorithm;
        }

        const int FIELD_TYPE_NAMESPACE = 0;
        const int FIELD_TYPE_TABLE = 1;
        const int FIELD_TYPE_KEY = 2;
        const int FIELD_TYPE_BIN = 3;
        const int FIELD_TYPE_DIGEST_RIPE160 = 4;
		const int FIELD_TYPE_GU_TID = 5;
		const int FIELD_TYPE_DIGEST_RIPE_ARRAY = 6;
		const int FIELD_TYPE_TRID = 7;	// user supplied transaction id, which is simply passed back
		const int FIELD_TYPE_SCAN_OPTIONS = 8;

        static private String utilArrayToString(Byte[] a)
        {
            String r = "";
            int i = 0;
            foreach (Byte b in a)
            {
                r += String.Format("{0:x2} ", b);
                if (i % 4 == 3) r += ':';
            }
            return (r);
        }

        /// <summary>
        /// Add output bytes to the collection. Return value is the new offset.
        /// Throws an array out of bounds if buffer is too small.
        /// </summary>
        static int makeOp(Op op, string name, object o, byte[] buf, int offset, IFormatter valueFromatter)
        {
            // byte - 1 byte
            // size - 7 bytes
            // op   - 1 byte
            // particle type - 1 byte
            // version - 1 byte
            // name_sz - 1 byte
            // name N
            // per-particle data

            int nameLen = StringToUtf8(name, buf, offset + 8);

            // only a true write needs the write data from object
            // the number 3 is the size *after* the size field
            int particle_offset = 8 + nameLen + offset;
            int field_sz = 4 + nameLen;
            ParticleType type = ParticleType.NULL;
            if (op == Op.WRITE || op == Op.ADD || op == Op.APPEND || op == Op.APPEND_EXT 
                || op == Op.APPEND_QUERY || op == Op.TOUCH )
            {
                field_sz += ObjectToParticleBytes(o, buf, particle_offset, out type, valueFromatter);
            }

            if (op == Op.APPEND || op == Op.APPEND_EXT || op == Op.APPEND_QUERY )
                type = ParticleType.APPEND_LIST;

            set_htonl((uint) field_sz, buf, ref offset);
            buf[offset] = (byte)op;
            buf[offset + 1] = (byte)type;
            buf[offset + 2] = 0; // version
            buf[offset + 3] = (byte)nameLen;

            return (field_sz + 4);
        }

        /// <summary>
        /// Sets 5 bytes of the field header.
        /// </summary>
        static int set_field_header(int type, long sz, byte[] buf, int offset)
        {
            //		Log.Debug(String.Format("set_field_header: sz "+sz+" offset "+offset));
            set_htonl((uint)(sz + 1), buf, ref offset);
            buf[offset] = (byte)type;
            offset++;
            //		Log.Debug(String.Format("bytes: "+buf[offset-1]+" "+buf[offset]+" "+buf[offset+1]+" "+buf[offset+2]));
            return (5);
        }

        public static int set_htonll(ulong v, byte[] buf, ref int offset)
        {
            byte[] va = System.BitConverter.GetBytes(v);
            if (System.BitConverter.IsLittleEndian)
                Array.Reverse(va);
            Array.Copy(va, 0, buf, offset, 8);
            offset += 8;
            return (8);
        }

        public static int set_htonll(ulong v, byte[] buf, int offset)
        {
            byte[] va = System.BitConverter.GetBytes(v);
            if (System.BitConverter.IsLittleEndian)
                Array.Reverse(va);
            Array.Copy(va, 0, buf, offset, 8);
            return (8);
        }

        public static int set_htonl(uint v, byte[] buf, ref int offset)
        {
            for (int i = 3; i >= 0; i--)
            {
                buf[offset + i] = (byte)(v & 0xff);
                v >>= 8;
            }
            offset += 4;
            return (4);
        }

        public static int set_htonl(uint v, byte[] buf, int offset)
        {
            for (int i = 3; i >= 0; i--)
            {
                buf[offset + i] = (byte)(v & 0xff);
                v >>= 8;
            }
            return (4);
        }

        public static int set_htons(ushort v, byte[] buf, int offset)
        {
            buf[offset] = (byte) (v >> 8);
            buf[offset + 1] = (byte) (v & 0xFF);
            return (2);
        }

        static long date_time_to_delta_seconds(DateTime t)
        {
            long delta = t.ToFileTime() - DateTime.Now.ToFileTime();
            if (delta <= 0) delta = 1;
            else delta /= 10000000;
            return ( delta );
        }

        /// <summary>
        /// A converter to fill a byte buffer with bytes.
        /// </summary>
        static int set_particle_ulong(ulong v, byte[] buf, int offset) 
        {
            byte[] va = System.BitConverter.GetBytes( v );
            if (System.BitConverter.IsLittleEndian)
                Array.Reverse(va);
            int sz;
            if (0 == (va[0] & 0x80))
            {
                buf[offset] = 0;
                sz = 1;
            }
            else
            {
                sz = 0;
            }
            Array.Copy(va, 0, buf, offset + sz, 8);
            return (8 + sz);
	    }

        static int set_particle_long(long v, byte[] buf, int offset)
        {
            if (v >= 0) return (set_particle_ulong((ulong)v, buf, offset));

            byte[] va = System.BitConverter.GetBytes( v );
            if (System.BitConverter.IsLittleEndian)
                Array.Reverse(va);
            Array.Copy(va, 0, buf, offset, 8);
            return (8);
        }

        static int set_particle_uint(uint v, byte[] buf, int offset)
        {
            for (int i = 3; i >= 0; i--)
            {
                buf[offset + i] = (byte)(v & 0xff);
                v >>= 8;
            }
            return (4);
	    }

        static int set_particle_int(int v, byte[] buf, int offset)
        {
            if (v >= 0) return (set_particle_uint((uint)v, buf, offset));
            // we know it's negative now
            if (v >= SByte.MinValue) return(set_particle_sbyte((sbyte)v, buf, offset));
            if (v >= Int16.MinValue) return(set_particle_short((short)v, buf, offset));

            for (int i = 3; i >= 0; i--)
            {
                buf[offset + i] = (byte)(v & 0xff);
                v >>= 8;
            }
            return (4);
        }

        static int set_particle_ushort(ushort v, byte[] buf, int offset)
        {
            buf[offset] = (byte)((v >> 8) & 0xff);
            buf[offset + 1] = (byte)(v & 0xff);
            return (2);
        }

        static int set_particle_short(short v, byte[] buf, int offset)
        {
            buf[offset] = (byte)((v >> 8) & 0xff);
            buf[offset + 1] = (byte)(v & 0xff);
            return (2);
        }

        static int set_particle_sbyte(sbyte v, byte[] buf, int offset)
        {
            buf[offset] = (byte)v;
            return (1);
        }

        static object get_integer(byte[] buf, ref int offset, int sz, ClLog log)
        {
            // short circuit 1
            if (sz == 1)
            {
                if (0 == (buf[offset] & 0x80))
                    return ((sbyte)buf[offset]);
                else
                    return (buf[offset]);
            }
            // general purpose
            bool signed = (0 == (buf[offset] & 0x80)) ? false : true;
            while ((sz > 0) && (buf[offset] == 0)) { offset++; sz--; }
            ulong acc = 0;

            for (int i = 0; i < sz ; i++)
            {
                acc <<= 8;
                acc |= buf[offset++];
            }
            switch (sz)
            {
                case 0:
                case 1:
                case 2:
                case 3:
                case 4:
                    if (signed) return (int)acc;
                    return ((uint)acc);
                case 5:
                case 6:
                case 7:
                case 8:
                    if (signed) return ((long)acc);
                    return ((ulong)acc);
                default:
                    log.Log(CitrusleafClient.LogLevel.WARNING, " illegal integer size: {0}", sz);
                    return (null);
            }
        }

        public static ulong get_ntohll(byte[] buf, int offset)
        {
            ulong a = 0;
            for (int i = 0; i < 8; i++)
            {
                a <<= 8;
                a |= buf[offset + i];
            }
            return (a);
        }

        static uint get_ntohl(byte[] buf, int offset)
        {
            uint a = 0;
            for (int i = 0; i < 4; i++)
            {
                a <<= 8;
                a |= buf[offset + i];
            }
            return (a);
        }

        static ushort get_ntohs(byte[] buf, int offset)
        {

            return ((ushort) ((buf[offset] << 8) + buf[offset + 1]));

        }

        public static byte[] makeDigest(string set, object key)
        {
            BinaryFormatter formatter = new BinaryFormatter();
            int capacity = ObjectToParticleBytesEstimate(key, formatter) + (set.Length * 2) + 1;
            byte[] buf = new byte[capacity]; 
            int set_len = StringToUtf8(set, buf, 0);

            ParticleType type;
            int key_len = ObjectToParticleBytes(key, buf, set_len + 1, out type, formatter) + 1; // type is one byte
            buf[set_len] = (byte)type;
            int count = key_len + set_len;

            byte[] hash_out = getHashAlgorithm().ComputeHash(buf, 0/*offset*/, count);

            /*
            if (log != null)
                log.Log(CitrusleafClient.LogLevel.VERBOSE,"makeDigest calculated digest {0} ", utilArrayToString(hash_out));
            */
            return (hash_out);
        }

        private static uint GetPartition(byte[] keyDigest)
        {
            uint p = BitConverter.ToUInt32(keyDigest, 0);
            return (p);
        }

        private static uint GetPartition(byte[] a1, int o1, int l1, byte[] a2, int o2, int l2, ClLog log)
        {

            byte[] hash_in = new byte[l1 + l2];
            Array.Copy(a1, o1, hash_in, 0, l1);
            Array.Copy(a2, o2, hash_in, l1, l2);

            byte[] hash_out = getHashAlgorithm().ComputeHash(hash_in);

            //log.Log(CitrusleafClient.LogLevel.VERBOSE, "get partition calculated digest {0} ", utilArrayToString(hash_out) );

            uint p = BitConverter.ToUInt32(hash_out, 0);
            return( p );
        }

        /// <summary>
        /// All requests have namespace, table, key, so make a byte array that includes all those fields.
        /// </summary>
        int MakeFieldsScan(byte[] buf, int offset, out uint partition, out ushort n_fields)
	    {
            // Make three headers.
            // 1 type = namespace
            // 3 len
            // [namespace bytes]
            // 1 type = table
            // 7 len
            // [table bytes]
            // 1 type = key
            // 7 len
            // 1 particle type
            // key particle bytes
            int initial_offset = offset;
            n_fields = 0;

            if (ns != null)
            {
                int ns_len = StringToUtf8(ns, buf, offset + 5);
                offset += set_field_header(FIELD_TYPE_NAMESPACE, ns_len, buf, offset) + ns_len;
                n_fields++;
            }

            if (set != null)
            {
                int table_len = StringToUtf8(set, buf, offset + 5);
                offset += set_field_header(FIELD_TYPE_TABLE, table_len, buf, offset) + table_len;
                n_fields++;
            }

            partition = uint.MaxValue;

            int scanOffset = offset + 5;
            // Server scan priority is "auto".  Scan will not stop if cluster breaks.
            buf[scanOffset] = 0;
            // Scan percent
            buf[scanOffset + 1] = 100;  
            offset += set_field_header(FIELD_TYPE_SCAN_OPTIONS, 2, buf, offset) + 2;
            n_fields++;
		    return(offset-initial_offset);
	    }

        int MakeFieldsDigest(byte[] buf, int offset, out uint partition, out ushort n_fields)
        {
            int initial_offset = offset;
            n_fields = 0;

            if (ns != null)
            {
                int ns_len = StringToUtf8(ns, buf, offset + 5);
                offset += set_field_header(FIELD_TYPE_NAMESPACE, ns_len, buf, offset) + ns_len;
                n_fields++;
            }

            if (set != null)
            {
                int table_len = StringToUtf8(set, buf, offset + 5);
                offset += set_field_header(FIELD_TYPE_TABLE, table_len, buf, offset) + table_len;
                n_fields++;
            }

            partition = uint.MaxValue;
            offset += set_field_header(FIELD_TYPE_DIGEST_RIPE160, keyDigest.Length, buf, offset);
            Array.Copy(keyDigest, 0, buf, offset, keyDigest.Length);
            offset += keyDigest.Length;
            n_fields++;

            partition = GetPartition(keyDigest);

            return (offset - initial_offset);
        }

        // Buffer instance variables and functions.
        public string ns;
        public string set;
        //public object key;
        public byte[] keyDigest;
        private ClLog log;

        private IList<string> readBins;
        private IDictionary<string, object> writeBins;
        private IDictionary<string, object> addBins;
        private IDictionary<string, object> appendBins;
        private IList<ClAppendExtBin> appendExtBins;
        private IList<ClAppendQueryBin> appendQueryBins;
        private bool readAll;
        private bool readIterate;
        private bool unique; 
        private bool binunique;
        private bool touch;
        private bool scan;

        private string write_single_name;
        private object write_single_value;
        private string read_single_name;
        private string add_single_name;
        private object add_single_value;
        private string append_single_name;
        private object append_single_value;

        CitrusleafWriteInfo winfo;

        private bool delete;

        byte[] msg_header_buf; // A 30 byte buffer used for reading just the response header
        byte[] msg_buf = null; // used to write the request, and read in the response
        int msg_write_len;

        public IFormatter valueFormatter { get; set; }  // serializer overwrite by  in by application
        
        public Buffer(ClLog log)
	    {
    	    ns = null;
    	    set = null;
    	    //key = null;
            keyDigest = null;
            this.log = log;
        	
    	    readBins = null;
    	    writeBins = null;
            addBins = null;
            appendBins = null;
            appendExtBins = null;
            appendQueryBins = null;
            write_single_name = null;
            write_single_value = null;
            read_single_name = null;
            add_single_name = null;
            add_single_value = null;
            append_single_name = null;
            append_single_value = null;

    	    readAll = false;
            readIterate = false;
            unique = false;
            binunique = false;
            touch = false;
            winfo = null;
            
            delete = false;

            msg_header_buf = null;
            msg_buf = null;
            valueFormatter = new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();
	    }

        /// <summary>
        /// Purpose of the reset function is to allow multiple reuse of an object before just leaving.
        /// </summary>
        public void Reset()
        {
    	    ns = null;
    	    set = null;
    	    //key = null;
            keyDigest = null;
        	
    	    readBins = null;
    	    writeBins = null;
            addBins = null;
            appendBins = null;
            appendExtBins = null;
            appendQueryBins = null;
            write_single_name = null;
            write_single_value = null;
            add_single_name = null;
            add_single_value = null;
            append_single_name = null;
            append_single_value = null;
            read_single_name = null;

    	    readAll = false;
            readIterate = false;
            unique = false;
            binunique = false;
            touch = false;
            winfo = null;
            
            delete = false;

            msg_header_buf = null;
            msg_buf = null;
        }

        /// <summary>
        /// Sets singleton initial values.
        /// Function must be expanded for secondary references
        /// </summary>
        public void Prepare(string ns, string set, object key)
        {
            this.ns = ns;
            this.set = set;
            this.keyDigest = Buffer.makeDigest(set, key);
        }

        public void Prepare(string ns, byte[] keyDigest)
        {
            this.ns = ns;
            this.keyDigest = keyDigest;
        }

        public void PrepareScan(string ns, string set)
        {
            this.ns = ns;
            this.set = set;
            this.keyDigest = null;
            this.scan = true;
        }

        /// <summary>
        /// on a read request - read this bin too.
        /// </summary>
        public void AddRead(string bin)
        {
            if (readBins != null)
            {
                readBins.Add(bin);
            }
            else
            {
                if (read_single_name == null)
                    read_single_name = bin;
                else
                {
                    readBins = new List<string>();
                    readBins.Add(read_single_name);
                    readBins.Add(bin);
                    read_single_name = null;
                }
            }
        }

        public void AddRead(IList<string> bins)
        {
            if (read_single_name != null)
            {
                readBins = bins;
                readBins.Add(read_single_name);
                read_single_name = null;
            }
            else
            {
                if (readBins == null)
                    readBins = bins;
                else
                {
                    foreach (string s in bins)
                        this.readBins.Add(s);
                }
            }
        }

        /// On a write request - write this bin too.
        /// </summary>
        public void AddWrite(IDictionary<string, object> writeBins)
        {
            if (this.writeBins == null)
            {
                this.writeBins = writeBins;
            }
            else
            {
                foreach (KeyValuePair<string, object> kvp in writeBins)
                    this.writeBins[kvp.Key] = kvp.Value;
            }
            // if there was a single, stash it in the dict
            if (write_single_name != null)
            {
                this.writeBins[write_single_name] = write_single_value;
                write_single_value = write_single_name = null;
            }
        }

        public void AddWrite(string bin, object o)
        {
            if (writeBins != null)
            {
                writeBins.Add(bin, o);
            }
            else
            {
                if (write_single_name == null)
                {
                    write_single_name = bin;
                    write_single_value = o;
                }
                else
                {
                    writeBins = new Dictionary<string, object>();
                    writeBins.Add(bin, o);
                    writeBins[write_single_name] = write_single_value;
                    write_single_name = null; 
                    write_single_value = null;
                }
            }
        }

        public void set_WriteInfo(CitrusleafWriteInfo winfo)
        {
            this.winfo = winfo;
        }

        public void set_Delete()
        {
            this.delete = true;
        }

        public void set_Unique()
        {
            this.unique = true;
        }

        public void set_BinUnique()
        {
            this.binunique = true;
        }

        public void set_ReadAll()
        {
            // if we previously had explict bins, throw them away
            if (readBins != null)
                readBins = null;
            if (read_single_name != null)
                read_single_name = null;
            readAll = true;
        }

        public void set_ReadIterate()
        {
            // if we previously had explict bins, throw them away
            if (readBins != null)
                readBins = null;
            if (read_single_name != null)
                read_single_name = null;
            readIterate = true;
        }

        public void set_Touch()
        {
            this.touch = true;
        }

        public void AddAdd(IDictionary<string, object> addBins)
        {
            // on a write request - write this bin too
            if (this.addBins == null)
            {
                this.addBins = addBins;
            }
            else
            {
                foreach (KeyValuePair<string, object> kvp in addBins)
                    this.addBins[kvp.Key] = kvp.Value;
            }
            // if there was a single, stash it in the dict
            if (add_single_name != null)
            {
                this.addBins[add_single_name] = add_single_value;
                add_single_value = add_single_name = null;
            }
        }

        public void AddAdd(string bin, object o)
        {
            if (addBins != null)
            {
                addBins.Add(bin, o);
            }
            else
            {
                if (add_single_name == null)
                {
                    add_single_name = bin;
                    add_single_value = o;
                }
                else
                {
                    addBins = new Dictionary<string, object>();
                    addBins.Add(bin, o);
                    addBins[add_single_name] = add_single_value;
                    add_single_name = null;
                    add_single_value = null;
                }
            }
        }

        public void AddAppend(IDictionary<string, object> appendBins)
        {
            // on a write request - write this bin too
            if (this.appendBins == null)
            {
                this.appendBins = appendBins;
            }
            else
            {
                foreach (KeyValuePair<string, object> kvp in appendBins)
                    this.appendBins[kvp.Key] = kvp.Value;
            }
            // if there was a single, stash it in the dict
            if (append_single_name != null)
            {
                this.appendBins[append_single_name] = append_single_value;
                append_single_value = append_single_name = null;
            }
        }

        public void AddAppend(string bin, object o)
        {
            if (appendBins != null)
            {
                appendBins.Add(bin, o);
            }
            else
            {
                if (append_single_name == null)
                {
                    append_single_name = bin;
                    append_single_value = o;
                }
                else
                {
                    appendBins = new Dictionary<string, object>();
                    appendBins.Add(bin, o);
                    appendBins[append_single_name] = append_single_value;
                    append_single_name = null;
                    append_single_value = null;
                }
            }
        }

        private class ClAppendExtBin
        {
            public string bin;
            public object value;
            public ClAppendWriteOptions write_options;
            public byte[] blob;

            public ClAppendExtBin(string bin, object value, ClAppendWriteOptions write_options)
            {
                this.bin = bin;
                this.value = value;
                this.write_options = write_options;
                this.blob = null;
            }

            public void pb_Serialize(IFormatter formatter)  // nb - formatter is used to format C# blobs. 
            {
#if APPEND_EXT_PROTOBUF_CSHARP
                AppendWriteOption.Builder write_opts_builder = new AppendWriteOption.Builder();
                int value_sz = ObjectToParticleBytesEstimate(this.value, formatter);
                byte[] byteArray = new byte[value_sz];
                ParticleType unused;
                ObjectToParticleBytes(this.value, byteArray, 0, out unused, formatter);
                Google.ProtocolBuffers.ByteString bs = Google.ProtocolBuffers.ByteString.CopyFrom(byteArray);  // not an obvious way not to copy here...
                write_opts_builder.SetValue(bs);
                if (this.write_options.tag != null) {
                    write_opts_builder.SetTag(this.write_options.tag);
                }
                if (this.write_options.creation_time != DateTime.MinValue) {
                    DateTime epochTime = new DateTime(1970,1,1);
                    TimeSpan t = this.write_options.creation_time - epochTime;
                    write_opts_builder.SetCreationTime((uint)t.TotalSeconds);
                }
                if (this.write_options.expiration != 0) {
                    write_opts_builder.SetTimeExpires(this.write_options.expiration);
                }

                AppendWriteOption write_option = write_opts_builder.BuildPartial();
                using (MemoryStream stream = new MemoryStream())
                {
                    write_option.WriteTo(stream);
                    this.blob = stream.ToArray();
                }
#elif APPEND_EXT_PROTOBUF_NET
                AppendWriteOption writeOption = new AppendWriteOption();
                int value_sz = ObjectToParticleBytesEstimate(this.value, formatter);
                byte[] byteArray = new byte[value_sz];
                ParticleType unused;
                ObjectToParticleBytes(this.value, byteArray, 0, out unused, formatter);
                writeOption.value = byteArray;
                if (this.write_options.tag != null)
                {
                    writeOption.tag = this.write_options.tag;
                }
                if (this.write_options.creation_time != DateTime.MinValue)
                {
                    DateTime epochTime = new DateTime(1970, 1, 1);
                    TimeSpan t = this.write_options.creation_time - epochTime;
                    writeOption.creationTime = (uint)t.TotalSeconds;
                }
                if (this.write_options.expiration != 0)
                {
                    writeOption.timeExpires = this.write_options.expiration;
                }
                using (MemoryStream stream = new MemoryStream())
                {
                    ProtoBuf.Serializer.Serialize<AppendWriteOption>(stream, writeOption);
                    this.blob = stream.ToArray();
                }
#endif
            }
        }

        private class ClAppendQueryBin
        {
            public string bin;
            public uint n_responses;
            public IList<ClAppendQueryFilter> filters;
            public byte[] blob;

            public ClAppendQueryBin(string bin, uint n_responses, IList<ClAppendQueryFilter> filters)
            {
                this.bin = bin;
                this.n_responses = n_responses;
                this.filters = filters;
                this.blob = null;
            }

            
            // serialize data for protobuf
            public void pb_Serialize()
            {
#if APPEND_EXT_PROTOBUF_CSHARP
                AppendQuery.Builder query_builder = AppendQuery.CreateBuilder();

                // set max results - simple
                query_builder.SetMaxResponses(this.n_responses);

                // query options - a little harder . Need to iterate through the filters
				if( this.filters != null ){
                    foreach (ClAppendQueryFilter filter in this.filters) {
                        // create filter
                        AppendQuery.Types.AppendQueryFilter.Builder query_filter_builder =
                            AppendQuery.Types.AppendQueryFilter.CreateBuilder();
                        DateTime epochTime = new DateTime(1970,1,1);
                        if (filter.start_time != DateTime.MinValue){
                            TimeSpan t = filter.start_time - epochTime;
                            query_filter_builder.SetStartTime((uint)t.TotalSeconds);
                        }
                        if (filter.end_time != DateTime.MinValue){
                            TimeSpan t = filter.end_time - epochTime;
                            query_filter_builder.SetEndTime((uint)t.TotalSeconds);
                        }
                        if (filter.tag != null) {
                            query_filter_builder.SetTag(filter.tag);
                        }
                        query_builder.AddFilter(query_filter_builder);
                    }
                }
                AppendQuery query = query_builder.BuildPartial();
                using (MemoryStream stream = new MemoryStream())
                {
                    query.WriteTo(stream);
                    this.blob = stream.ToArray();
                }
#elif APPEND_EXT_PROTOBUF_NET
                AppendQuery appendQuery = new AppendQuery();

                // set max results - simple
                appendQuery.maxResponses = this.n_responses;

                // query options - a little harder . Need to iterate through the filters
                if (this.filters != null)
                {
                    foreach (ClAppendQueryFilter filter in this.filters)
                    {
                        // create filter
                        AppendQuery.AppendQueryFilter queryFilter = new AppendQuery.AppendQueryFilter();
                        DateTime epochTime = new DateTime(1970, 1, 1);
                        if (filter.start_time != DateTime.MinValue)
                        {
                            TimeSpan t = filter.start_time - epochTime;
                            queryFilter.startTime = (uint)t.TotalSeconds;
                        }
                        if (filter.end_time != DateTime.MinValue)
                        {
                            TimeSpan t = filter.end_time - epochTime;
                            queryFilter.endTime = (uint)t.TotalSeconds;
                        }
                        if (filter.tag != null)
                        {
                            queryFilter.tag = filter.tag;
                        }
                        appendQuery.filter.Add(queryFilter);
                    }
                }
                using (MemoryStream stream = new MemoryStream())
                {
                    ProtoBuf.Serializer.Serialize<AppendQuery>(stream, appendQuery);
                    this.blob = stream.ToArray();
                }
#endif
            }
        }

        public void AddAppendExt(string bin, object o, ClAppendWriteOptions append_write_options)
        {
            ClAppendExtBin append_ext_bin = new ClAppendExtBin(bin, o, append_write_options);

            if (this.appendExtBins == null)
            {
                appendExtBins = new List<ClAppendExtBin>();
            }
            appendExtBins.Add(append_ext_bin);
        }

        public void AddAppendQuery(string bin, uint n_responses, IList<ClAppendQueryFilter> filters)
        {
            ClAppendQueryBin append_query_bin = new ClAppendQueryBin(bin, n_responses, filters);

            if (this.appendQueryBins == null)
            {
                this.appendQueryBins = new List<ClAppendQueryBin>();
            }

            appendQueryBins.Add(append_query_bin);
        }

        /// <summary>
        /// Do all the buffer preparation that makes it ready to read or write.
        /// </summary>
        public CitrusleafResult Ready(out uint partition) 
        {
            int n_ops = 0;
            partition = 0;
            // Estimate buffer length
            int len = 22 + 8;
            len += 2 * (4 + 1 + 32); // two fixed-size headers

            //if (key != null)
            //    len += 4 + 1 + ObjectToParticleBytesEstimate(key, new BinaryFormatter());
            if (keyDigest != null)
                len += 4 + 1 + keyDigest.Length;
            
            if (writeBins != null)
            {
                foreach (object o in writeBins.Values)
                {
                    int rv = ObjectToParticleBytesEstimate(o, valueFormatter );
                    if (rv < 0)
                    {
                        log.Log(CitrusleafClient.LogLevel.WARNING,"could not serialize object {0}", o.ToString());
                        return (CitrusleafResult.CLIENT_ERROR);
                    }
                    len += rv;
                    len += 8 + 34; // op header + max name
                }
                n_ops += writeBins.Count;
            }
            if (write_single_name != null)
            {
                int rv = ObjectToParticleBytesEstimate(write_single_value, valueFormatter);
                if (rv < 0)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING,"could not serialize object {0}", write_single_value.ToString());
                    return (CitrusleafResult.CLIENT_ERROR);
                }
                len += rv;
                len += 6 + 34; // op header + max name
                n_ops++;
            }

            if (addBins != null)
            {
                foreach (object o in addBins.Values)
                {
                    int rv = ObjectToParticleBytesEstimate(o,valueFormatter);
                    if (rv < 0)
                    {
                        log.Log(CitrusleafClient.LogLevel.WARNING, "could not serialize object {0}", o.ToString());
                        return (CitrusleafResult.CLIENT_ERROR);
                    }
                    len += rv;
                    len += 8 + 34; // op header + max name
                }
                n_ops += addBins.Count;
            }
            if (add_single_name != null)
            {
                int rv = ObjectToParticleBytesEstimate(add_single_value,valueFormatter);
                if (rv < 0)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, "could not serialize object {0}", write_single_value.ToString());
                    return (CitrusleafResult.CLIENT_ERROR);
                }
                len += rv;
                len += 6 + 34; // op header + max name
                n_ops++;
            }

            if (appendBins != null)
            {
                foreach (object o in appendBins.Values)
                {
                    int rv = ObjectToParticleBytesEstimate(o,valueFormatter);
                    if (rv < 0)
                    {
                        log.Log(CitrusleafClient.LogLevel.WARNING, "could not serialize object {0}", o.ToString());
                        return (CitrusleafResult.CLIENT_ERROR);
                    }
                    len += rv;
                    len += 8 + 34; // op header + max name
                }
                n_ops += appendBins.Count;
            }
            if (append_single_name != null)
            {
                int rv = ObjectToParticleBytesEstimate(append_single_value,valueFormatter);
                if (rv < 0)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, "could not serialize object {0}", write_single_value.ToString());
                    return (CitrusleafResult.CLIENT_ERROR);
                }
                len += rv;
                len += 6 + 34; // op header + max name
                n_ops++;
            }

            if (appendQueryBins != null)
            {
                foreach (ClAppendQueryBin bin in appendQueryBins)
                {
                    if (bin.blob == null){
                        bin.pb_Serialize();
                    }

                    len += bin.blob.Length;

                    len += 8 + 34; // op header + max name
                }
                n_ops += appendQueryBins.Count;
            }

            if (appendExtBins != null)
            {
                foreach (ClAppendExtBin bin in appendExtBins)
                {
                    if (bin.blob == null) {
                        bin.pb_Serialize(valueFormatter);
                    }

                    len += bin.blob.Length;
                    len += 8 + 34; // op header + max name
                }
                n_ops += appendExtBins.Count;
            }
            if (touch == true)
            {
                n_ops++;
            }

            // Continue the estimation - include read bins if available
            if (readBins != null)
            {
                len += readBins.Count * (6 + 34);
                n_ops += readBins.Count;
            }
            if (read_single_name != null)
            {
                len += (6 + 34);
                n_ops++;
            }

            msg_buf = new byte[len];
    		
            // Skip ahead and lay in the fields
            ushort n_fields;
            int offset = (22 + 8);

            if (scan)
            {
                offset += MakeFieldsScan(msg_buf, offset, out partition, out n_fields);
            }
            else
            {
                offset += MakeFieldsDigest(msg_buf, offset, out partition, out n_fields);
            }
            //if (keyDigest != null)
            //    offset += MakeFieldsDigest(msg_buf, offset, out partition, out n_fields);
            //else
            //    offset += MakeFields(msg_buf, offset, out partition, out n_fields);

            // now the read ops
            byte info1 = 0;
            byte info2 = 0;
            byte info3 = 0;
            if (readBins != null)
            {
                foreach (string s in readBins)
                {
				    offset += makeOp(Op.READ, s, null, msg_buf, offset, valueFormatter);
			    }
			    info1 |= CITRUSLEAF_INFO1_READ;
		    }
            else if (read_single_name != null)
            {
                offset += makeOp(Op.READ, read_single_name, null, msg_buf, offset, valueFormatter);
                info1 |= CITRUSLEAF_INFO1_READ;
            }
            else if (readAll)
            {
                info1 |= CITRUSLEAF_INFO1_GET_ALL;
                info1 |= CITRUSLEAF_INFO1_READ;
            }
            else if (readIterate)
            {
                info1 |= CITRUSLEAF_INFO1_READ;
            }

            // and the append queries...
            if (appendQueryBins != null)
            {
                foreach (ClAppendQueryBin bin in appendQueryBins)
                {
                    if (bin.blob == null) {
                        bin.pb_Serialize();
                    }

                    offset += makeOp(Op.APPEND_QUERY, bin.bin, bin.blob, msg_buf, offset, valueFormatter);
			    }
			    info1 |= CITRUSLEAF_INFO1_READ;
		    }

    		
            if (writeBins != null || write_single_name != null)
            {
                if (writeBins != null)
                {
                    foreach (KeyValuePair<string, object> kvp in writeBins)
                    {
                        offset += makeOp(Op.WRITE, kvp.Key, kvp.Value, msg_buf, offset, valueFormatter);
                    }
                }
                if (write_single_name != null)
                {
                    offset += makeOp(Op.WRITE, write_single_name, write_single_value, msg_buf, offset, valueFormatter);
                }

                info2 |= CITRUSLEAF_INFO2_WRITE;
                if (unique)
                {
                    info2 |= CITRUSLEAF_INFO2_WRITE_UNIQUE;
                }
                if (binunique)
                {
                    info2 |= CITRUSLEAF_INFO2_WRITE_BINUNIQUE;
                }
		    }
            else if (touch == true)
            {
                info2 |= CITRUSLEAF_INFO2_WRITE;
                offset += makeOp(Op.TOUCH, "", null, msg_buf, offset, valueFormatter);
            }

            if (addBins != null || add_single_name != null)
            {
                if (addBins != null)
                {
                    foreach (KeyValuePair<string, object> kvp in addBins)
                    {
                        offset += makeOp(Op.ADD, kvp.Key, kvp.Value, msg_buf, offset, valueFormatter);
                    }
                }
                if (add_single_name != null)
                {
                    offset += makeOp(Op.ADD, add_single_name, add_single_value, msg_buf, offset, valueFormatter);
                }

                info2 |= CITRUSLEAF_INFO2_WRITE;
            }

            if (appendBins != null || append_single_name != null)
            {
                if (appendBins != null)
                {
                    foreach (KeyValuePair<string, object> kvp in appendBins)
                    {
                        offset += makeOp(Op.APPEND, kvp.Key, kvp.Value, msg_buf, offset, valueFormatter);
                    }
                }
                if (append_single_name != null)
                {
                    offset += makeOp(Op.APPEND, append_single_name, append_single_value, msg_buf, offset, valueFormatter);
                }

                info2 |= CITRUSLEAF_INFO2_WRITE;
            }

            if (appendExtBins != null)
            {
                foreach (ClAppendExtBin bin in appendExtBins)
                {
                    if (bin.blob == null)
                    {
                        bin.pb_Serialize(valueFormatter);
                    }

                    offset += makeOp(Op.APPEND_EXT, bin.bin, bin.blob, msg_buf, offset, valueFormatter);
                }

                info2 |= CITRUSLEAF_INFO2_WRITE;
            }
            msg_write_len = offset;

            uint generation;
            if (winfo != null && winfo.GenerationSet())
            {
                info2 |= CITRUSLEAF_INFO2_GENERATION;
                generation = winfo.Generation;
            }
            else generation = 0;
    		
            if (delete)
            {
                info2 |= CITRUSLEAF_INFO2_WRITE;
                info2 |= CITRUSLEAF_INFO2_DELETE;
		    }

		    // Make the actual header
            // magic constants like this are annoying, but the '2' is the protocol version
            // and the '3' is the message within the protocol
		    ulong sz =  (ulong) ((2L << 56) | (3L << 48)) + (ulong) (msg_write_len - 8);
		    set_htonll(sz, msg_buf, 0);
		    msg_buf[8] = 22;
		    msg_buf[9] = info1;
            msg_buf[10] = info2;
            msg_buf[11] = info3;
            msg_buf[12] = 0; // unused
            msg_buf[13] = 0; // result_code
            set_htonl(generation, msg_buf, 14);
            if (winfo != null)
            {
                set_htonl((uint)winfo.object_expiration_seconds, msg_buf, 18);
            }
            else
            {
                msg_buf[18] = msg_buf[19] = msg_buf[20] = msg_buf[21] = 0;
            }
    		
		    // number of fields: Table. key. + number of read and write bins
		    set_htons((ushort)n_fields, msg_buf, 26);

		    set_htons((ushort)n_ops,msg_buf,28);
    		
		    // DEBUG
		    // dump_buf("write msg ",msg_buf,(int)msg_write_len);
            return (CitrusleafResult.OK);		
	    }

        CitrusleafResult ParseResult(byte[] msg, ref int offset, int n_fields, int n_ops, int resultCode, out string nsname, out string setname, out IDictionary<string, object> values, out byte[] digest)
        {
    	    CitrusleafResult r;
            nsname = null;
            setname = null;
            values = null;
            digest = null;

            switch (resultCode)
            {
		        case 0:
		            r = CitrusleafResult.OK;
				    break;
			    case 1:
                    r = CitrusleafResult.UNKNOWN_ERROR;
                    break;
			    case 2:
				    r = CitrusleafResult.KEY_NOT_FOUND_ERROR;
				    break;
			    case 3:
				    r = CitrusleafResult.GENERATION_ERROR;
				    break;
		        case 4:
                    r = CitrusleafResult.PARAMETER_ERROR;
                    break;
                case 5:
                    r = CitrusleafResult.KEY_FOUND_ERROR;
                    break;
                case 6:
                    r = CitrusleafResult.BIN_FOUND_ERROR;
                    break;
                case 7:
                    r = CitrusleafResult.CLUSTER_KEY_MISMATCH;
                    break;
                case 8:
                    r = CitrusleafResult.SERVER_MEM_ERROR;
                    break;
                case 9:
                    r = CitrusleafResult.TIMEOUT;
                    break;
                case 10:
                    r = CitrusleafResult.NO_XDS;
                    break;
                case 11:
                    r = CitrusleafResult.SERVER_NOT_AVAILABLE;
                    break;
                case 12:
                    r = CitrusleafResult.BIN_TYPE_ERROR;
                    break;
                case 13:
                    r = CitrusleafResult.RECORD_TOO_BIG;
                    break;
                case 14:
                    r = CitrusleafResult.KEY_BUSY;
                    break;
                default:
                    log.Log(CitrusleafClient.LogLevel.WARNING, "new error type {0}", resultCode);
                    r = CitrusleafResult.UNKNOWN_ERROR;
                    break;
            }

            if (n_fields > 0)
            {

                for (int i = 0; i < n_fields; i++)
                {
                    int field_sz = (int)get_ntohl(msg, offset);
                    byte type = msg[offset + 4];

                    if (type == FIELD_TYPE_DIGEST_RIPE160)
                    {
                        digest = new byte[field_sz - 1];
                        Array.Copy(msg, offset + 5, digest, 0, field_sz - 1);
                    }
                    else if (type == FIELD_TYPE_NAMESPACE)
                    {
                        nsname = Encoding.UTF8.GetString(msg, offset + 5, field_sz - 1);
                    }
                    else if (type == FIELD_TYPE_TABLE)
                    {
                        setname = Encoding.UTF8.GetString(msg, offset + 5, field_sz - 1);
                    }
                    offset += 4 + field_sz;
                    log.Log(CitrusleafClient.LogLevel.VERBOSE,"recv message: field set: type " + type + " size " + field_sz);
                }

		    }

            // put the result into a dictionary
            if (n_ops > 0)
            {

                for (int i = 0; i < n_ops; i++)
                {

                    int op_sz = (int)get_ntohl(msg, offset);
                    byte op = msg[offset + 4];
                    byte particle_type = msg[offset + 5];
                    byte version = msg[offset + 6];
                    byte name_sz = msg[offset + 7];
                    string name = Utf8ToString(msg, offset + 8, name_sz);

                    /*
                    Console.WriteLine("op size {0} op {1} part-type {2} name_sz {3} name {4}\n",
                    					op_sz, op, particle_type, name_sz, name);
                    */
                    object value = BytesToParticle(particle_type, msg, offset + 8 + name_sz, op_sz - (4 + name_sz), log, valueFormatter );
                    offset += 4 + op_sz;
                    if (value != null)
                    {
                        if (values == null)
                            values = new Dictionary<string, object>(n_ops);
                        values[name] = value;
                    }
                }
            }
            return (r);
        }

        public CitrusleafResult Put(Socket socket)
        {
            int pos = 0;
            try
            {
                while (pos < msg_write_len)
                {
                    pos += socket.Send(msg_buf, pos, msg_write_len - pos, SocketFlags.None);
                }
            }
            catch (SocketException ex)
            {
                if (ex.SocketErrorCode == SocketError.TimedOut)
                    return CitrusleafResult.TIMEOUT;

                log.Log(CitrusleafClient.LogLevel.INFO, " put data: internal error: exception {0}", ex.ToString());
                return CitrusleafResult.SERVER_ERROR;
            }
            return CitrusleafResult.OK;
        }

        public CitrusleafResult Get(Socket socket, out IDictionary<string, object> values, out uint generation)
        {
            try
            {
                return GetData(socket, out values, out generation);
            }
            catch (SocketException ex)
            {
                generation = 0;
                values = null;

                if (ex.SocketErrorCode == SocketError.TimedOut)
                    return CitrusleafResult.TIMEOUT;

                log.Log(CitrusleafClient.LogLevel.INFO, " GET: socket Receive Error: {0} {1} {2}. Returning SERVER_ERROR ",
                    ex.SocketErrorCode.ToString(), socket.LocalEndPoint.ToString(), socket.RemoteEndPoint.ToString());
                return CitrusleafResult.SERVER_ERROR;
            }
        }

        private CitrusleafResult GetData(Socket socket, out IDictionary<string, object> values, out uint generation)
        {
            generation = 0;
            values = null;

            if (msg_header_buf == null) 
                msg_header_buf = new byte[30];

            if (socket.ReceiveTimeout > 0)
            {
                // Check if data is available for reading.
                // Poll is used because the timeout value is respected under 500ms.
                // The Receive method does not timeout until after 500ms.
                if (! socket.Poll(socket.ReceiveTimeout * 1000, SelectMode.SelectRead))
                {
                    throw new SocketException((int)SocketError.TimedOut);
                }
             }

            int rlen = 0;
            while (rlen < 30)
            {
                int rlen_now = socket.Receive(msg_header_buf, rlen, 30 - rlen, SocketFlags.None);

                if (rlen_now <= 0)
                {
                    log.Log(CitrusleafClient.LogLevel.INFO, " GET: received {0} bytes {1} {2} . Returning SERVER_ERROR ",
                        rlen_now, socket.LocalEndPoint.ToString(), socket.RemoteEndPoint.ToString());
                    return CitrusleafResult.SERVER_ERROR;
                }
                rlen += rlen_now;
            }

            // Debug
            // dump_buf("read header", msg_header_buf, 0);

            ulong sz = get_ntohll(msg_header_buf, 0);
            byte version = (byte)(((int)(sz >> 56)) & 0xff);
            byte type = (byte)(((int)(sz >> 48)) & 0xff);
            byte h_len = msg_header_buf[8];
            byte info1 = msg_header_buf[9];
            byte info2 = msg_header_buf[10];
            byte info3 = msg_header_buf[11];
            byte unused = msg_header_buf[12];
            byte resultCode = msg_header_buf[13];
            generation = get_ntohl(msg_header_buf, 14);
            uint record_ttl = get_ntohl(msg_header_buf, 18);
            uint transaction_ttl = get_ntohl(msg_header_buf, 22);
            //Console.WriteLine("record ttl-ms is {0}", record_ttl);

            if (version != 2)
            {
                log.Log(CitrusleafClient.LogLevel.WARNING, " parse get: version is incorrect, should be 2 is " + version);
                return (CitrusleafResult.SERVER_ERROR);
            }
            if (type != 3)
            {
                log.Log(CitrusleafClient.LogLevel.WARNING, " parse get: type is incorrect, should be 3 is " + version);
                return (CitrusleafResult.SERVER_ERROR);
            }

            int n_fields = get_ntohs(msg_header_buf, 26); // May get the setname back
            int n_ops = get_ntohs(msg_header_buf, 28);
            int n_bytes = ((int)(sz & 0xFFFFFFFFFFFFL)) - h_len;

            //		System.out.printf("read header: version %d hlen %d info %x result %d nops %d nbytes %d\n",
            //				version, h_len, info, resultCode, n_ops, n_bytes);

            byte[] local_msg_buf = null;
            if (n_bytes != 0)
            {
                if (msg_buf == null)
                {
                    msg_buf = new byte[1024 * 24];
                }

                if (msg_buf.Length < n_bytes)
                {
                    local_msg_buf = new byte[n_bytes];
                }
                else
                {
                    local_msg_buf = msg_buf;
                }

                // readit
                rlen = 0;
                while (rlen < n_bytes)
                {
                    int rlen_now = socket.Receive(local_msg_buf, rlen, n_bytes - rlen, SocketFlags.None);

                    if (rlen_now <= 0)
                    {
                        log.Log(CitrusleafClient.LogLevel.INFO, "GET (2): body received 0 bytes. Returning SERVER_ERROR ");
                        return CitrusleafResult.SERVER_ERROR;
                    }
                    rlen += rlen_now;
                }
                //dump_buf("read msg body", local_msg_buf, n_bytes);
            }

            // parseit
            string nsname;
            string setname;
            byte[] digest;
            int offset = 0;
            CitrusleafResult r = ParseResult(local_msg_buf, ref offset, n_fields, n_ops, resultCode, out nsname, out setname, out values, out digest);

            // take a look at what was returned and what was asked for.
            // if not all the bins are there, return the bin not found
            if ((values != null) && (values.Count == 0)) values = null;
            if (r == CitrusleafResult.OK)
            {
                if (readBins != null)
                {
                    if (values == null) r = CitrusleafResult.BIN_NOT_FOUND_ERROR;
                    else if (values.Count != readBins.Count) r = CitrusleafResult.BIN_NOT_FOUND_ERROR;
                }
                else if (read_single_name != null)
                {
                    if (values == null) r = CitrusleafResult.BIN_NOT_FOUND_ERROR;
                    else if (values.Count != 1) r = CitrusleafResult.BIN_NOT_FOUND_ERROR;
                }
            }
            return (r);
        }

        public CitrusleafResult GetScan(ScanThread thread, Socket socket, CitrusleafClient.IScanCallback callback)
        {
            try
            {
                return GetScanData(thread, socket, callback);
            }
            catch (SocketException ex)
            {
                if (ex.SocketErrorCode == SocketError.TimedOut)
                    return CitrusleafResult.TIMEOUT;

                log.Log(CitrusleafClient.LogLevel.INFO, "GETMulti: socket Receive Error {0}. Returning SERVER_ERROR ", ex.SocketErrorCode.ToString());
                return CitrusleafResult.SERVER_ERROR;
            }
        }

        public CitrusleafResult GetScanData(ScanThread thread, Socket socket, CitrusleafClient.IScanCallback callback)
        {
            CitrusleafResult r = CitrusleafResult.OK;

            if (msg_header_buf == null) 
                msg_header_buf = new byte[30];

            do
            {
                // log.Log(CitrusleafClient.LogLevel.VERBOSE,"Get multi: found object ");

                // Parse out the CL header first
                int rlen = 0;
                while (rlen < 8)
                {
					if (!thread.Running)
					{
						return CitrusleafResult.SCAN_TERMINATED;
					}

					int rlen_now = socket.Receive(msg_header_buf, rlen, 8 - rlen, SocketFlags.None);
                    
                    if (rlen_now <= 0)
                    {
                        log.Log(CitrusleafClient.LogLevel.INFO, " GetMulti: received 0 bytes. Returning SERVER_ERROR ");
                        return CitrusleafResult.SERVER_ERROR;
                    }
                    rlen += rlen_now;
                }
                ulong sz = get_ntohll(msg_header_buf, 0);
                byte version = (byte)(((int)(sz >> 56)) & 0xff);
                byte type = (byte)(((int)(sz >> 48)) & 0xff);
                int n_bytes = (int)(sz & 0xFFFFFFFFFFFFL);

                // validate some message headers
                if (version != 2)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, " parse get: version is incorrect, should be 2 is " + version);
                    return (CitrusleafResult.SERVER_ERROR);
                }
                if (type != 3)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, " parse get: type is incorrect, should be 3 is " + version);
                    return (CitrusleafResult.SERVER_ERROR);
                }

                byte[] buf = null;
                if (n_bytes != 0)
                {
                    if (msg_buf == null)
                    {
                        msg_buf = new byte[1024 * 24];
                    }

                    if (msg_buf.Length < n_bytes)
                    {
                        buf = new byte[n_bytes];
                    }
                    else
                    {
                        buf = msg_buf;
                    }

                    // readit
                    rlen = 0;
                    while (rlen < n_bytes)
                    {
						if (!thread.Running)
						{
							return CitrusleafResult.SCAN_TERMINATED;
						}

						int rlen_now = socket.Receive(buf, rlen, n_bytes - rlen, SocketFlags.None);

                        if (rlen_now <= 0)
                        {
                            log.Log(CitrusleafClient.LogLevel.INFO, " GetMulti (2): body received 0 bytes. Returning SERVER_ERROR ");
                            return (CitrusleafResult.SERVER_ERROR);
                        }
                        rlen += rlen_now;
                    }

                    // then loop over the contained as messages
                    int offset = 0;
                    do
                    {
                        uint generation = 0;
                        IDictionary<string, object> values = null;

                        // Debug
                        // dump_buf("read header", msg_header_buf, 0);

                        byte h_len = buf[offset];
                        byte info1 = buf[offset+1];
                        byte info2 = buf[offset+2];
                        byte info3 = buf[offset+3];
                        byte unused = buf[offset+4];
                        byte resultCode = buf[offset+5];
                        generation = get_ntohl(buf, offset + 6);
                        uint record_ttl = get_ntohl(buf, offset + 10);
                        uint transaction_ttl = get_ntohl(buf, offset + 14);

                        int n_fields = get_ntohs(buf, offset + 18); 
                        int n_ops = get_ntohs(buf, offset + 20);
                        offset += 22;

                        if (h_len != 22)
                        {
                            return (CitrusleafResult.SERVER_ERROR);
                        }

                        // check for last message
                        if ((info3 & CITRUSLEAF_INFO3_LAST) == CITRUSLEAF_INFO3_LAST)
                        {
                            //		System.out.printf("read header: version %d hlen %d info %x result %d nops %d nbytes %d\n",
                            //				version, h_len, info, resultCode, n_ops, n_bytes);
                            return (r);
                        }


                        // parseit
                        string nsname;
                        string setname;
                        byte[] digest;
                        r = ParseResult(buf, ref offset, n_fields, n_ops, resultCode, out nsname, out setname, out values, out digest);
                        if (r != CitrusleafResult.OK)
                        {
                            return (r);
                        }

						if (!thread.Running)
						{
							return CitrusleafResult.SCAN_TERMINATED;
						}
						
						//dump_buf("read msg body", local_msg_buf, n_bytes);
                        if (!callback.ScanRecordCallback(nsname, setname, digest, values, generation, record_ttl))
                        {
                            // The callback returned false which indicates the scan should be terminated.
                            return (CitrusleafResult.SCAN_TERMINATED);
                        }

                    } while (offset < n_bytes);

                }

            } while (true);
        }

        /// <summary>
        /// Copy object into buffer, return the type and the new offset. 
        /// </summary>
        static int ObjectToParticleBytes(object o, byte[] buf, int offset, out ParticleType type, IFormatter formatter)
        {
            // And besides array segemnts of bytes, we have straight byte arrays
            if (o is byte[])
            {
                byte[] tmpByteArray = o as byte[];
                type = ParticleType.BLOB;
                Array.Copy(tmpByteArray, 0, buf, offset, tmpByteArray.Length);
                return (tmpByteArray.Length);
            }
            // raw byte arrays are common and quick, so special case them first
            if (o is ArraySegment<byte>)
            {
                ArraySegment<byte> asb = (ArraySegment<byte>) o;
                type = ParticleType.BLOB;
                Array.Copy(asb.Array, asb.Offset, buf, offset, asb.Count);
                return (asb.Count);
            }

            TypeCode code = ( (o == null) ? TypeCode.Empty : Type.GetTypeCode(o.GetType()) );

            byte[] data;

            switch (code)
            {
                case TypeCode.Empty:
                    type = ParticleType.NULL;
                    return (0);

                case TypeCode.String:
                    type = ParticleType.STRING;
                    data = Encoding.UTF8.GetBytes((string)o);
                    Array.Copy(data, 0, buf, offset, data.Length );
                    return (data.Length);

                case TypeCode.Boolean:
                    type = ParticleType.INTEGER;
                    buf[offset] = ((bool)o) == true ? (byte)1 : (byte)0;
                    return (1);

                case TypeCode.Int16:
                    type = ParticleType.INTEGER;
                    return (set_particle_long(Convert.ToInt64(o), buf, offset));

                case TypeCode.Int32:
                    type = ParticleType.INTEGER;
                    return (set_particle_long(Convert.ToInt64(o), buf, offset));

                case TypeCode.Int64:
                    type = ParticleType.INTEGER;
                    return (set_particle_long(Convert.ToInt64(o), buf, offset));

                case TypeCode.UInt16:
                    type = ParticleType.INTEGER;
                    return (set_particle_long(Convert.ToInt64(o), buf, offset));

                case TypeCode.UInt32:
                    type = ParticleType.INTEGER;
                    return (set_particle_long(Convert.ToInt64(o), buf, offset));

                case TypeCode.UInt64:
                    type = ParticleType.INTEGER;
                    return (set_particle_long(Convert.ToInt64(o), buf, offset));

                case TypeCode.Byte:
                    type = ParticleType.INTEGER;
                    buf[offset] = (byte)o;
                    return (1);

                case TypeCode.SByte:
                    type = ParticleType.INTEGER;
                    return(set_particle_sbyte((sbyte)o, buf, offset));

                case TypeCode.Char:
                case TypeCode.DateTime:
                case TypeCode.Double:
                case TypeCode.Single:
                default:
                    type = ParticleType.CSHARP_BLOB;
                    using (MemoryStream ms = new MemoryStream())
                    {
                        formatter.Serialize(ms, o);

                        code = TypeCode.Object;
                        int bsz = buf.Length - (offset + 1);
                        ms.Seek(0, 0);
                        int rsz = ms.Read(buf, offset, bsz);
                        return (rsz);
                    }
            }
        }

        /// <summary>
        /// Estimate the size of the bytes.
        /// Optimization: it would be nice to stash and recover the actual object in cases
        /// where we had to really build the object.
        /// </summary>
        static int ObjectToParticleBytesEstimate(object o,IFormatter formatter)
        {
            // raw byte arrays are common and quick, so special case them first
            if (o is byte[]) return (((byte[])o).Length + 2);
            // raw byte arrays are common and quick, so special case them first
            if (o is ArraySegment<byte>) return(((ArraySegment<byte>)o).Count + 2);

            TypeCode code = ((o == null) ? TypeCode.Empty : Type.GetTypeCode(o.GetType()));

            switch (code)
            {
                case TypeCode.Empty:
                    return (0);

                case TypeCode.String:
                    return (((string)o).Length * 2);

                case TypeCode.Boolean:
                case TypeCode.Byte:
                case TypeCode.SByte:
                    return (2);

                case TypeCode.UInt16:
                case TypeCode.Int16:
                case TypeCode.Int32:
                case TypeCode.UInt32:
                case TypeCode.Int64:
                case TypeCode.UInt64:
                    return (10);

                case TypeCode.Char:
                case TypeCode.DateTime:
                case TypeCode.Double:
                case TypeCode.Single:
                default:
                    using (MemoryStream ms = new MemoryStream())
                    {
                        formatter.Serialize(ms, o);
                        return ((int)ms.Length + 1);
                    }
            }
        }

        /// <summary>
        /// Binary Format of the object is: type (dict or list)
        /// </summary>
        static object ParseList(byte[] buf, int offset, int len, ClLog log, IFormatter valueFormatter)
        {
            int limit = offset + len;
            int n_items = (int) get_ntohl(buf, offset);
            offset += 4;
            List<object> r = new List<object>(n_items);

            while (offset < limit)
            {
                int sz = (int) get_ntohl(buf, offset);
                offset += 4;
                int type = buf[offset];
                offset++;

                r.Add(BytesToParticle(type, buf, offset, sz, log,valueFormatter));
                offset += sz;

            }

            return (r);
        }

        static object ParseDict(byte[] buf, int offset, int len, ClLog log, IFormatter valueFormatter)
        {
            int limit = offset + len;
            int n_items = (int)get_ntohl(buf, offset);
            offset += 4;
            Dictionary<object,object> d = new Dictionary<object,object>(n_items);

            while (offset < limit)
            {
                // read out the key
                int sz = (int)get_ntohl(buf, offset);
                offset += 4;
                int type = buf[offset];
                offset++;

                object key = BytesToParticle(type, buf, offset, len, log, new BinaryFormatter());
                offset += sz;

                // read out the value
                sz = (int)get_ntohl(buf, offset);
                offset += 4;
                type = buf[offset];
                offset++;

                object value = BytesToParticle(type, buf, offset, len, log, valueFormatter);
                offset += sz;

                d.Add(key, value);

            }

            return (d);
        }

        static object BytesToParticle(int type, byte[] buf, int offset, int len, ClLog log, IFormatter valueFormatter)
        {
            if (type == (int) ParticleType.STRING)
            {
                return Encoding.UTF8.GetString(buf, offset, len);
            }
            else if (type == (int) ParticleType.INTEGER)
            {
                return( get_integer(buf, ref offset, len, log) );
            }
            else if (type == (int)ParticleType.BLOB)
            {
                byte[] r = new byte[len];
                Array.Copy(buf, offset, r, 0, len);
                return(r);
            }
            else if (type == (int) ParticleType.CSHARP_BLOB)
            {
                try
                {
                    MemoryStream ms = new MemoryStream();
                    ms.Write(buf, offset, len);
                    ms.Seek(0, 0);
                    object o = valueFormatter.Deserialize(ms);
                    return (o);
                }
                catch (SerializationException) 
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, "unable to parse blob size {0}",len);
                    return (null);
                }
            }
            else if (type == (int)ParticleType.LIST)
            {
                return ParseList(buf, offset, len, log, valueFormatter);
            }
            else if (type == (int)ParticleType.DICT)
            {
                return ParseDict(buf, offset, len, log, valueFormatter);
            }

            if (type == (int)ParticleType.NULL)
            {
                return (null);
            }

            log.Log(CitrusleafClient.LogLevel.WARNING, "bytesToParticle: Unsupported type {0}", type);
            return(null);
        }

        /// <summary>
        /// This is a ported-from-java UTF8 converter. 
        /// Returns the number of bytes this transalated into.
        /// </summary>
        public static int StringToUtf8_custom(string s, byte[] buf, ref int offset, ClLog log)
        {
            int l = s.Length;
            int start_offset = offset;
            for (int i = 0; i < l; i++)
            {
                char c = s[i];
                if (c < 0x7f)
                {
                    buf[offset] = (byte) c;
                    offset++;
                }
                else if (c < 0x07FF)
                {
                    buf[offset] = (byte)(0xC0 | (c >> 6));
                    buf[offset + 1] = (byte)(0x80 | (c & 0x5f));
                    offset += 2;
                }
                else
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, "internal error in StringToUtf8 routine");
                }
            }
            return(offset - start_offset);
        }

        static int StringToUtf8(string s, byte[] buf, ref int offset)
        {
            byte[] data = Encoding.UTF8.GetBytes(s);
            Array.Copy(data, 0, buf, offset, data.Length);
            return (data.Length);
        }

        static int StringToUtf8(string s, byte[] buf, int offset)
        {
            byte[] data = Encoding.UTF8.GetBytes(s);
            Array.Copy(data, 0, buf, offset, data.Length);
            return (data.Length);
        }

        /// <summary>
        /// Return a worst-case size.
        /// </summary>
        public static int StringToUtf8_len(string s)
        {
            return (s.Length * 2);
        }

        /// <summary>
        /// This is my custom UTF8 converter. Under java it runs faster than the native versions.
        /// Currently not used for C#
        /// </summary>
        public static string Utf8ToString_custom(byte[] buf, int offset, int length)
        {
            StringBuilder sb = new StringBuilder(length);
            int limit = offset + length;

            while (offset < limit)
            {

                if ((buf[offset] & 0x80) == 0)
                { // 1 byte
                    char c = (char)buf[offset];
                    sb.Append(c);
                    offset++;
                }
                else if ((buf[offset] & 0x1f) == 0xC0)
                { // 2 bytes
                    char c = (char)(((buf[offset] & 0x1f) << 6) | (buf[offset + 1] & 0x3f));
                    sb.Append(c);
                    offset += 2;
                }
            }
            return (sb.ToString());

        }

        public static string Utf8ToString(byte[] buf, int offset, int length)
        {
            return (System.Text.Encoding.UTF8.GetString(buf, offset, length));
        }

        private static void dump_buf(string info, byte[] buf, int limit)
        {
    	    Console.WriteLine("dump buffer: "+info+" size "+buf.Length+" limit "+limit);
    	    if (limit == 0) limit = buf.Length;
            for (int i = 0; i < limit; i++)
            {
    		    Console.Write(String.Format(" {0,2:X}", buf[i]));
    		    if (i % 16 == 7)
    			    Console.Write(":");
    		    if (i % 16 == 15)
    			    Console.WriteLine();
    	    }
    	    Console.WriteLine();
        }       
    }	
}
	
