using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace net.citrusleaf
{
    /// <summary>
    /// This client handles administrative tasks needed to manage communication 
    /// with server nodes. The client knows the optimal cluster node for each transaction, 
    /// handles retries, and manages cluster reconfiguration issues in a way that is 
    /// transparent to the application.  The client converts database requests into the wire
    /// protocol and sends each request to a server node in the cluster.
    /// </summary>
    public class CitrusleafClient : IDisposable
    {
        private Cluster asc = null;
        private ClLog log;
        private Boolean disposed = false;
        private Settings defaultSettings;
        private string defaultNamespace;
        private string defaultSet;
        private string defaultBin;

        /// <summary>
        /// Construct a CirusleafClient object and initialize logging framework.
        /// </summary>
        public CitrusleafClient()
        {
            log = new ClLog();
            defaultSettings = new Settings();
            CreateValueFormatter = () => new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();
        }

        //----------------------------------------------------------------------------------------
        // Database defaults.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Default settings used for managing client connections.
        /// </summary>
        public Settings DefaultSettings
        {
            get { return defaultSettings; }
            set { defaultSettings = (value ?? new Settings()); }
        }

        /// <summary>
        /// Default namespace.  If the namespace is not specified in the API call, DefaultNamespace will be used.
        /// The default should be modified before multiple client library threads are created.
        /// </summary>
        public string DefaultNamespace
        {
            get { return (defaultNamespace); }
            set { defaultNamespace = value; }
        }

        /// <summary>
        /// Default set name.  If the set name is not specified in the API call, DefaultSet will be used.
        /// The default should be modified before multiple client library threads are created.
        /// </summary>
        public string DefaultSet
        {
            get { return (defaultSet); }
            set { defaultSet = value; }
        }

        /// <summary>
        /// Default bin name.  If the bin name is not specified in the API call, DefaultBin will be used.
        /// Set to empty string if server nodes are configured as "single-bin".
        /// The default should be modified before multiple client library threads are created.
        /// </summary>
        public string DefaultBin
        {
            get { return (defaultBin); }
            set { defaultBin = value; }
        }

        //----------------------------------------------------------------------------------------
        // Connection.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Add a host to the client's cluster map.  The following actions occur upon the first
        /// invocation of this method:
        /// <list type="bullet">
        ///   <item><description>Create new cluster map.</description></item>
        ///   <item><description>Add host to cluster map.</description></item>
        ///   <item><description>Connect to host server.</description></item>
        ///   <item><description>Request host's list of other nodes in the cluster.</description></item>
        ///   <item><description>Add these nodes to the cluster map.</description></item>
        /// </list> 
        /// Further invocations will add hosts to the cluster map if they don't already exist. In most
        /// cases, only one AddHost() call is necessary.  When this call succeeds, the client 
        /// is ready to process database requests.
        /// </summary>
        /// <param name="host">Host name.</param>
        /// <param name="port">Host port.</param>
        /// <param name="ipMapping">Optional host name/IP address mapping. Use in cases where DNS lookups are not desirable.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AddHost(string host, int port, Dictionary<string,string> ipMapping = null)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already dispose");
            }

            // already have a cluster, just adding more hosts to it
            if (asc != null)
            {
                log.Log(CitrusleafClient.LogLevel.INFO, "addHost: adding additional host to existing shared cluster");
                return asc.AddHost(host, port);
            }

            // phase 1, if we already have one in the list, return it
            if ((asc=Cluster.tryGetClusterInstance(host, port, log)) != null)
                return CitrusleafResult.OK;

            // phase 2, if not available, create one. the AddHost call takes a long time, so it is out of the critical section
            // Someone else may come in and create one too and insert it into our dictionary, so this may be a throw away
            Cluster newCluster;
            try
            {
                newCluster = new Cluster(defaultSettings, log, ipMapping, host, port, 200);
            }
            catch (Exception ex)
            {
                log.Log(CitrusleafClient.LogLevel.WARNING, ex.Message);
                return CitrusleafResult.SERVER_ERROR;
            }

            // phase 3, either add this new cluster or try to get one that's been created
            asc = Cluster.getOrAdd(newCluster, host, port, log);

            // phase 4, dispose the redundant copy in case someone else won
            if (asc != newCluster)
            {
                log.Log(CitrusleafClient.LogLevel.VERBOSE, "getCluster: disposing redundant cluster for {0} {1}", host,port);
                newCluster.Dispose();
            }
            return CitrusleafResult.OK;
        }

        //----------------------------------------------------------------------------------------
        // Close and cleanup.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Close client connections to servers.
        /// </summary>
        public void Dispose()
        {
            if (disposed == false)
            {
                if (asc != null)
                {
                    Cluster.removeClusterInstance(asc, log);
                    asc = null;
                }
                GC.SuppressFinalize(this);
            }
            disposed = true;
        }

        /// <summary>
        /// Close client connections to servers. 
        /// </summary>
        ~CitrusleafClient()
        {
            if (!disposed)
            {
                if (asc != null)
                {
                    Cluster.removeClusterInstance(asc, log);
                    asc = null;
                }
            }
            disposed = true;
        }

        //----------------------------------------------------------------------------------------
        // Database operations using default namespace, set, and bin
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Set bin value for specified key.  
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Set(object key, object value, CitrusleafWriteInfo writeInfo)
        {
            return (Set(DefaultNamespace, DefaultSet, key, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key. The set will fail if the key already exists.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUnique(object key, object value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<string, object> values = new Dictionary<string, object>();
            values.Add(DefaultBin, value);
            return (SetInternal(DefaultNamespace, DefaultSet, key, values, true, false, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key. The set will fail if the default bin name already exists.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetBinUnique(object key, object value, CitrusleafWriteInfo writeInfo)
        {
            return (SetBinUnique(DefaultNamespace, DefaultSet, key, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Get bin value and record generation for specified key.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Returned bin value.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Transaction timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Get(object key, out object value, out uint generation, int timeoutMillis)
        {
            return (Get(DefaultNamespace, DefaultSet, key, DefaultBin, out value, out generation, timeoutMillis));
        }

        /// <summary>
        /// Batch mulitple get requests into a single call.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keys">Collection of keys.</param>
        /// <param name="values">Returned positionally matched bin values.</param>
        /// <param name="generations">Returned positionally matched generations. Generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Transaction timeout in milliseconds.</param>
        /// <returns>A result status for every key.</returns>        
        public CitrusleafResult[] BatchGet(ICollection<object> keys, out object[] values, out uint[] generations, int timeoutMillis)
        {
            return (BatchGet(DefaultNamespace, DefaultSet, keys, DefaultBin, out values, out generations, timeoutMillis));
        }

        /// <summary>
        /// Reset record time to expiration without changing its value for specified key.
        /// The record generation count will also be incremented.
        /// This method returns KEY_NOT_FOUND_ERROR if the record does not exist.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="expires">Seconds until expiration.</param>
        /// <param name="timeoutMillis">Transaction timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Touch(object key, int expires, int timeoutMillis)
        {
            return (Touch(DefaultNamespace, DefaultSet, key, expires, timeoutMillis));
        }

        /// <summary>
        /// Used for integer values only.
        /// Add value to existing bin value for specified key.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Add(object key, object value, CitrusleafWriteInfo writeInfo)
        {
            return (Add(DefaultNamespace, DefaultSet, key, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Delete a record for specified key. 
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Delete(object key, CitrusleafWriteInfo writeInfo)
        {
            return (Delete(DefaultNamespace, DefaultSet, key, writeInfo));
        }

        //----------------------------------------------------------------------------------------
        // Database operations using namespace and set as method arguments.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Set record bin value for specified key and bin name.
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Set(string ns, string set, object key, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<string, object> values = new Dictionary<string, object>();
            values.Add(bin, value);
            return SetInternal(ns, set, key, values, false, false, writeInfo);
        }

        /// <summary>
        /// Set record bin value for specified key and bin name.
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Set(string ns, string set, object key, string bin, string value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<string, object> values = new Dictionary<string, object>();
            values.Add(bin, value);
            return SetInternal(ns, set, key, values, false, false, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key and bin names.
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Set(string ns, string set, object key, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            return SetInternal(ns, set, key, values, false, false, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key and bin names. 
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Set(string ns, string set, object key, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // shortcut for now - just zip the two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (Set(ns, set, key, values_dict, writeInfo));
        }

        /// <summary>
        /// Set record bin value for specified key and bin name. The set will fail if the 
        /// key already exists.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUnique(string ns, string set, object key, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<string, object> values = new Dictionary<string, object>();
            values.Add(bin, value);
            return SetUnique(ns, set, key, values, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key and bin names. The set will fail if the 
        /// key already exists.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="values">Bin name/values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUnique(string ns, string set, object key, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            return SetInternal(ns, set, key, values, true, false, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key and bin names. The set will fail if the 
        /// key already exists.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUnique(string ns, string set, object key, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // shortcut for now - just zip the two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (SetUnique(ns, set, key, values_dict, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key and bin name. The set will fail if the bin name already exists.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetBinUnique(string ns, string set, object key, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<string, object> values = new Dictionary<string, object>();
            values.Add(bin, value);
            return SetBinUnique(ns, set, key, values, writeInfo);
        }

        /// <summary>
        /// Set bin values for specified key and bin names. The set will fail if the bin name 
        /// already exists.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetBinUnique(string ns, string set, object key, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            return SetInternal(ns, set, key, values, false, true, writeInfo);
        }

        /// <summary>
        /// Set bin values for specified key and bin names. The set will fail if the bin name 
        /// already exists.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetBinUnique(string ns, string set, object key, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // put the two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (SetBinUnique(ns, set, key, values_dict, writeInfo));
        }

        /// <summary>
        /// Get bin value and generation for specified key and bin name.  The get will timeout if the server 
        /// does not respond within the specified time period. 
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name filter.</param>
        /// <param name="value">Returned bin value.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Get(string ns, string set, object key, string bin, out object value, out uint generation, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // make sure there is a key
            value = null;
            generation = 0;

            if (key == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddRead(bin);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            IDictionary<string, object> values;
            r = Do(partition, false, b, timeoutMillis, out values, out generation);

            if (r != CitrusleafResult.OK) return (r);
            if (values != null)
            {
                if (values.ContainsKey(bin))
                {
                    value = values[bin];
                }
                else
                {
                    return CitrusleafResult.BIN_NOT_FOUND_ERROR;
                }
            }

            return r;
        }

        /// <summary>
        /// Get dictionary of bin name/value pairs for specified key and list of bin names.
        /// The get will timeout if the server does not respond within the specified time period. 
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bins">Bin names filter.</param>
        /// <param name="values">Returned bin name/value pairs.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Get(string ns, string set, object key, IList<string> bins, out IDictionary<string, object> values, out uint generation, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            values = null;
            generation = 0;

            if (key == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddRead(bins);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            return (Do(partition, false, b, timeoutMillis, out values, out generation));
        }

        /// <summary>
        /// Get all bins for specified key. The record generation count is also returned.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="values">Returned bin name/value pairs.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetAll(string ns, string set, object key, out IDictionary<string, object> values, out uint generation, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            values = null;
            generation = 0;
            uint partition;

            if (key == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.set_ReadAll();
            CitrusleafResult r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            return (Do(partition, false, b, timeoutMillis, out values, out generation));
        }

        /// <summary>
        /// Get bin value for specified key and bin name.
        /// If the record exists, reset record time to expiration and increment record generation count.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name filter.</param>
        /// <param name="value">Returned bin value.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Client request timeout in milliseconds.</param>
        /// <param name="object_expiry_seconds">Record expiration time period in seconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetWithTouch(string ns, string set, object key, string bin, out object value, out uint generation, int timeoutMillis, int object_expiry_seconds)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // make sure there is a key
            value = null;
            generation = 0;

            if (key == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddRead(bin);

            CitrusleafWriteInfo writeInfo = new CitrusleafWriteInfo(timeoutMillis);
            writeInfo.object_expiration_seconds = object_expiry_seconds;
            b.set_WriteInfo(writeInfo);
            b.set_Touch();

            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            IDictionary<string, object> values;
            r = Do(partition, false, b, timeoutMillis, out values, out generation);

            if (r != CitrusleafResult.OK) return (r);
            if (values != null)
            {
                if (values.ContainsKey(bin))
                {
                    value = values[bin];
                }
                else
                {
                    return CitrusleafResult.BIN_NOT_FOUND_ERROR;
                }
            }

            return r;
        }

        /// <summary>
        /// Batch multiple get requests into a single call.  Return a list of dictionaries, each dictionary 
        /// containing a collection of bin name/value pairs.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="keys">Collection of keys.</param>
        /// <param name="bins">Bin names filter.</param>
        /// <param name="values">Returned positionally matched dictionaries.  Each dictionary contains a list of bin name/values pairs.</param>
        /// <param name="generations">Returned positionally matched generations.</param>
        /// <param name="timeoutMillis">Client request timeout in milliseconds.</param>
        /// <returns>A result status for every key.</returns>
        public CitrusleafResult[] BatchGet(string ns, string set, ICollection<object> keys, IList<string> bins, out IDictionary<string, object>[] values, out uint[] generations, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            int req_sz = keys.Count;

            CitrusleafResult[] r_a = new CitrusleafResult[req_sz];
            generations = new uint[req_sz];
            values = new IDictionary<string, object>[req_sz];

            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();

            int i = 0;
            foreach (object key in keys)
            {
                b.Prepare(ns, set, key);
                b.AddRead(bins);
                uint partition;
                r_a[i] = b.Ready(out partition);
                if (r_a[i] != CitrusleafResult.OK) continue;

                r_a[i] = Do(partition, false, b, timeoutMillis, out values[i], out generations[i]);

                b.Reset();
                i++;
            }
            return (r_a);
        }

        /// <summary>
        /// Batch multiple get requests into a single call.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="keys">Positionally matched keys.</param>
        /// <param name="bin">Bin name filter.</param>
        /// <param name="values">Returned positionally matched bin values.</param>
        /// <param name="generations">Returned positionally matched generations.</param>
        /// <param name="timeoutMillis">Client request timeout in milliseconds.</param>
        /// <returns>A result status for every key.</returns>
        public CitrusleafResult[] BatchGet(string ns, string set, ICollection<object> keys, string bin, out object[] values, out uint[] generations, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            int req_sz = keys.Count;

            CitrusleafResult[] r_a = new CitrusleafResult[req_sz];
            generations = new uint[req_sz];
            values = new object[req_sz];

            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();

            int i = 0;
            foreach (object key in keys)
            {
                b.Prepare(ns, set, key);
                b.AddRead(bin);
                uint partition;
                r_a[i] = b.Ready(out partition);
                if (r_a[i] != CitrusleafResult.OK) continue;

                IDictionary<string, object> valueList;
                r_a[i] = Do(partition, false, b, timeoutMillis, out valueList, out generations[i]);
                values[i] = valueList[bin];

                b.Reset();
                i++;
            }
            return (r_a);
        }

        /// <summary>
        /// Batch multiple get requests into a single call. 
        /// Return a list of dictionaries, each dictionary containing a collection of bin name/value pairs. 
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="keys">Collection of keys.</param>
        /// <param name="values">Returned positionally matched dictionaries.  Each dictionary contains a list of bin name/values pairs.</param>
        /// <param name="generations">Returned positionally matched generations.</param>
        /// <param name="timeoutMillis">Client request timeout in milliseconds.</param>
        /// <returns>A result status for every key.</returns>
        public CitrusleafResult[] BatchGetAll(string ns, string set, ICollection<object> keys, out IDictionary<string, object>[] values, out uint[] generations, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already dispose");
            }
            int req_sz = keys.Count;

            CitrusleafResult[] r_a = new CitrusleafResult[req_sz];
            generations = new uint[req_sz];
            values = new IDictionary<string, object>[req_sz];

            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();

            int i = 0;
            foreach (object key in keys)
            {
                b.Prepare(ns, set, key);
                b.set_ReadAll();
                uint partition;
                r_a[i] = b.Ready(out partition);
                if (r_a[i] != CitrusleafResult.OK) continue;

                IDictionary<string, object> valueList;
                r_a[i] = Do(partition, false, b, timeoutMillis, out valueList, out generations[i]);
                values[i] = valueList;

                b.Reset();
                i++;
            }
            return (r_a);
        }

        /// <summary>
        /// Reset record time to expiration without changing its value for specified key.
        /// The record generation count will also be incremented.
        /// This method returns KEY_NOT_FOUND_ERROR if the record does not exist.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="expiration_seconds">Seconds until expiration.</param>
        /// <param name="timeoutMillis">Client request timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Touch(string ns, string set, object key, int expiration_seconds, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already dispose");
            }
            uint partition;

            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            CitrusleafWriteInfo writeInfo = new CitrusleafWriteInfo(timeoutMillis);
            writeInfo.object_expiration_seconds = expiration_seconds;
            b.set_WriteInfo(writeInfo);
            b.set_Touch();

            CitrusleafResult r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            IDictionary<string, object> values;
            return (Do(partition, false, b, timeoutMillis, out values, out generation));
        }

        /// <summary>
        /// Batch multiple touch requests into a single call.
        /// For each key, reset record time to expiration without changing its value
        /// and increment record generation count.
        /// This method returns KEY_NOT_FOUND_ERROR for each key that does not exist.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="keys">Collection of keys.</param>
        /// <param name="expiration">Seconds until expiration for all keys.</param>
        /// <param name="timeoutMillis">Client request timeout in milliseconds.</param>
        /// <returns>A result status for every key.</returns>
        public CitrusleafResult[] BatchTouch(string ns, string set, ICollection<object> keys, int expiration, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already dispose");
            }
            int req_sz = keys.Count;

            CitrusleafResult[] r_a = new CitrusleafResult[req_sz];

            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();

            int i = 0;
            foreach (object key in keys)
            {
                b.Prepare(ns, set, key);

                CitrusleafWriteInfo writeInfo = new CitrusleafWriteInfo(timeoutMillis);
                writeInfo.object_expiration_seconds = expiration;
                b.set_WriteInfo(writeInfo);
                b.set_Touch();

                uint partition;
                r_a[i] = b.Ready(out partition);
                if (r_a[i] != CitrusleafResult.OK) continue;

                IDictionary<string, object> values;
                uint generation;
                r_a[i] = Do(partition, false, b, timeoutMillis, out values, out generation);

                b.Reset();
                i++;
            }
            return (r_a);
        }

        /// <summary>
        /// Used for integer values only. Add value to existing bin value for specified key.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Add(string ns, string set, object key, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already dispose");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddAdd(bin, value);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);

            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            IDictionary<string, object> values;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Used for integer values only.
        /// Add value to existing bin value for each bin for specified key.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Add(string ns, string set, object key, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddAdd(values);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);

            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Used for integer values only.
        /// Add value to existing bin value for each bin for specified key.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Add(string ns, string set, object key, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // put two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (Add(ns, set, key, values_dict, writeInfo));
        }

        /// <summary>
        /// Delete record for specified key.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Delete(string ns, string set, object key, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // make sure there is a key
            if (key == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            b.set_Delete();
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            IDictionary<string, object> values;
            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        //----------------------------------------------------------------------------------------
        // Database operations using digests.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Generate digest from a combination of key and set name attributes.
        /// </summary>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="log">Citrusleaf logger used for debugging.  Set to null for normal operation.</param>
        /// <returns>Unique ID generated from the key and set name.</returns>
        public static byte[] ComputeDigest(string set, object key, ClLog log)
        {
            return (Buffer.makeDigest(set, key));
        }

        /// <summary>
        /// Set bin value for specified key digest.  
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetDigest(byte[] keyDigest, object value, CitrusleafWriteInfo writeInfo)
        {
            return (SetDigest(DefaultNamespace, keyDigest, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key and bin name digest.  
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetDigest(string ns, byte[] keyDigest, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<string, object> values = new Dictionary<string, object>();
            values.Add(bin, value);
            return SetDigestInternal(ns, keyDigest, values, false, false, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key digest and bin names.
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetDigest(string ns, byte[] keyDigest, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            return SetDigestInternal(ns, keyDigest, values, false, false, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key digest and bin names. 
        /// If the key or bin does not exist, it will be created.
        /// If the bin exists its value is replaced.  Others bins are ignored.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetDigest(string ns, byte[] keyDigest, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // shortcut for now - just zip the two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (SetDigest(ns, keyDigest, values_dict, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key digest. The set will fail if the key already exists.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUniqueDigest(byte[] keyDigest, object value, CitrusleafWriteInfo writeInfo)
        {
            return (SetUniqueDigest(DefaultNamespace, keyDigest, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key and bin name digest. The set will fail if the key already exists.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUniqueDigest(string ns, byte[] keyDigest, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<string, object> values = new Dictionary<string, object>();
            values.Add(bin, value);
            return SetUniqueDigest(ns, keyDigest, values, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key digest and bin names. The set will fail if the 
        /// key digest already exists.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="values">Bin name/values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUniqueDigest(string ns, byte[] keyDigest, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            return SetDigestInternal(ns, keyDigest, values, true, false, writeInfo);
        }

        /// <summary>
        /// Set record bin values for specified key and bin names. The set will fail if the 
        /// key already exists.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetUniqueDigest(string ns, byte[] keyDigest, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // put the list together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (SetUniqueDigest(ns, keyDigest, values_dict, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key digest. The set will fail if the bin name already exists.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetBinUniqueDigest(byte[] keyDigest, object value, CitrusleafWriteInfo writeInfo)
        {
            return (SetBinUniqueDigest(DefaultNamespace, keyDigest, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Set bin value for specified key and bin name digest and bin name. The set will fail if the bin name exists.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult SetBinUniqueDigest(string ns, byte[] keyDigest, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            Dictionary<String, object> values = new Dictionary<string, object>();
            values.Add(bin, value);
            return SetDigestInternal(ns, keyDigest, values, false, true, writeInfo);
        }

        /// <summary>
        /// Get bin value and record generation for specified key digest.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="value">Returned bin value.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Transaction timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetDigest(byte[] keyDigest, out object value, out uint generation, int timeoutMillis)
        {
            return (GetDigest(DefaultNamespace, keyDigest, DefaultBin, out value, out generation, timeoutMillis));
        }

        /// <summary>
        /// Get bin value and generation for specified key digest and bin name.  The get will timeout if 
        /// the server does not respond within the specified time period. 
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bin">Bin name filter.</param>
        /// <param name="value">Returned bin value.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetDigest(string ns, byte[] keyDigest, string bin, out object value, out uint generation, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            value = null;
            generation = 0;

            if (keyDigest == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.AddRead(bin);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            IDictionary<string, object> values;
            r = Do(partition, false, b, timeoutMillis, out values, out generation);
            if (r != CitrusleafResult.OK) return (r);

            if (values != null)
            {
                if (values.ContainsKey(bin))
                {
                    value = values[bin];
                }
                else
                {
                    return CitrusleafResult.BIN_NOT_FOUND_ERROR;
                }
            }

            return r;
        }

        /// <summary>
        /// Get dictionary of bin name/value pairs for specified key digest and list of bin names.
        /// The get will timeout if the server does not respond within the specified time period. 
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bins">Bin names filter.</param>
        /// <param name="values">Returned bin name/value pairs.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetDigest(string ns, byte[] keyDigest, IList<string> bins, out IDictionary<string, object> values, out uint generation, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            values = null;
            generation = 0;

            if (keyDigest == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.AddRead(bins);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            return (Do(partition, false, b, timeoutMillis, out values, out generation));
        }

        /// <summary>
        /// Get all bins for specified key digest. The record generation count is also returned.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Not used.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="values">Returned bin name/value pairs.</param>
        /// <param name="generation">Returned generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetAllDigest(string ns, string set, byte[] keyDigest, out IDictionary<string, object> values, out uint generation, int timeoutMillis)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            values = null;
            generation = 0;
            uint partition;

            if (keyDigest == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.set_ReadAll();
            CitrusleafResult r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            return (Do(partition, false, b, timeoutMillis, out values, out generation));
        }

        /// <summary>
        /// Reset record time to expiration without changing its value for specified key digest.
        /// The record generation count will also be incremented.
        /// This method returns KEY_NOT_FOUND_ERROR if the record does not exist.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="expires">Seconds until expiration.</param>
        /// <param name="timeoutMillis">Transaction timeout in milliseconds.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult TouchDigest(byte[] keyDigest, int expires, int timeoutMillis)
        {
			return (TouchDigest(DefaultNamespace, keyDigest, expires, timeoutMillis));
		}

		/// <summary>
		/// Reset record time to expiration without changing its value for specified key digest.
		/// The record generation count will also be incremented.
		/// This method returns KEY_NOT_FOUND_ERROR if the record does not exist.
		/// Non-Digest database methods send the normal key and set name to the server which
		/// converts them to a digest. Digest methods send the digest directly.
		/// The following properties must be set before using this method. 
		/// <list type="bullet">
		///   <item><description>DefaultNamespace</description></item>
		///   <item><description>DefaultSet</description></item>
		/// </list> 
		/// </summary>
		/// <param name="ns">Namespace.</param>
		/// <param name="keyDigest">Unique ID generated from the key and set name.</param>
		/// <param name="expires">Seconds until expiration.</param>
		/// <param name="timeoutMillis">Transaction timeout in milliseconds.</param>
		/// <returns>Result status.</returns>
		public CitrusleafResult TouchDigest(string ns, byte[] keyDigest, int expires, int timeoutMillis)
		{
			if (disposed)
			{
				throw new ObjectDisposedException("citrusleaf client already dispose");
			}
			uint partition;

			Buffer b = new Buffer(log);
			b.valueFormatter = CreateValueFormatter();
			b.Prepare(ns, keyDigest);
			CitrusleafWriteInfo writeInfo = new CitrusleafWriteInfo(timeoutMillis);
			writeInfo.object_expiration_seconds = expires;
			b.set_WriteInfo(writeInfo);
			b.set_Touch();

			CitrusleafResult r = b.Ready(out partition);
			if (r != CitrusleafResult.OK) return (r);

			uint generation;
			IDictionary<string, object> values;
			return (Do(partition, false, b, timeoutMillis, out values, out generation));
		}
		
		/// <summary>
        /// Used for integer values only. 
        /// Add value to existing bin value for specified key digest.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AddDigest(byte[] keyDigest, object value, CitrusleafWriteInfo writeInfo)
        {
            return (AddDigest(DefaultNamespace, keyDigest, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Used for integer values only. 
        /// Add value to existing bin value for specified key digest.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AddDigest(string ns, byte[] keyDigest, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.AddAdd(bin, value);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);

            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            IDictionary<string, object> values;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Used for integer values only.
        /// Add value to existing bin value for each bin for specified key digest.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AddDigest(string ns, byte[] keyDigest, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.AddAdd(values);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);

            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Used for integer values only.
        /// Add value to existing bin value for each bin for specified key digest.
        /// If the record or bin does not exist, it will be created with the specified value.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AddDigest(string ns, byte[] keyDigest, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // put the two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (AddDigest(ns, keyDigest, values_dict, writeInfo));
        }

        /// <summary>
        /// Delete a row for specified key digest. 
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult DeleteDigest(byte[] keyDigest, CitrusleafWriteInfo writeInfo)
        {
            return (DeleteDigest(DefaultNamespace, keyDigest, writeInfo));
        }

        /// <summary>
        /// Delete record for specified key digest.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult DeleteDigest(string ns, byte[] keyDigest, CitrusleafWriteInfo writeInfo)
        {
            CitrusleafResult r;
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // make sure there is a key
            if (keyDigest == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            b.set_Delete();
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            IDictionary<string, object> values;
            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        //----------------------------------------------------------------------------------------
        // Server Status.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Get all server status data. Please see handbook for detailed statistics returned.
        /// </summary>
        /// <param name="host">Server hostname.</param>
        /// <param name="port">Server port.</param>
        /// <param name="values">Returned server attribute name/value pairs.</param>
        /// <param name="log">Citrusleaf logger used for debugging. Set to null for normal operation.</param>
        /// <param name="addressMappings">Optional dictionary of hostname/ip address mappings.</param>
        /// <returns>Result status.</returns>
        public static CitrusleafResult Info(string host, int port, out IDictionary<string, string> values, ClLog log, Dictionary<string, string> addressMappings = null)
        {
            IPEndPoint[] ipes = Cluster.Resolve(host, port, log);
            CitrusleafResult r = CitrusleafResult.SERVER_ERROR;
            values = null;
            if (ipes != null)
            {
                foreach (IPEndPoint ipe in ipes)
                {
                    r = net.citrusleaf.Info.Get(ipe, null, out values, log);
                    if (r == CitrusleafResult.OK) return (r);
                }
            }
            return (r);
        }

        /// <summary>
        /// Get server status attribute for specified name.
        /// </summary>
        /// <param name="host">Server hostname.</param>
        /// <param name="port">Server port.</param>
        /// <param name="name">Server attribute name filter.</param>
        /// <param name="value">Returned server attribute value.</param>
        /// <param name="log">Citrusleaf logger used for debugging. Set to null for normal operation.</param>
        /// <param name="addressMappings">Optional dictionary of hostname/ip address mappings.</param>
        /// <returns>Result status.</returns>
        public static CitrusleafResult Info(string host, int port, string name, out string value, ClLog log, Dictionary<string, string> addressMappings = null)
        {
            IPEndPoint[] ipes = Cluster.Resolve(host, port, log);
            CitrusleafResult r = CitrusleafResult.SERVER_ERROR;
            value = null;
            if (ipes != null)
            {
                foreach (IPEndPoint ipe in ipes)
                {
                    r = net.citrusleaf.Info.Get(ipe, name, out value, log);
                    if (r == CitrusleafResult.OK) return (r);
                }
            }
            return (r);
        }

        /// <summary>
        /// Get server status attributes for specified list of names.
        /// </summary>
        /// <param name="host">Server hostname.</param>
        /// <param name="port">Server port.</param>
        /// <param name="names">Server attribute names filter.</param>
        /// <param name="values">Returned attribute name/value pairs.</param>
        /// <param name="log">Citrusleaf logger used for debugging. Set to null for normal operation.</param>
        /// <param name="addressMappings">Optional dictionary of hostname/ip address mappings.</param>
        /// <returns>Result status.</returns>
        public static CitrusleafResult Info(string host, int port, IList<string> names, out IDictionary<string, string> values, ClLog log, Dictionary<string, string> addressMappings = null)
        {
            IPEndPoint[] ipes = Cluster.Resolve(host, port, log);
            CitrusleafResult r = CitrusleafResult.SERVER_ERROR;
            values = null;
            if (ipes != null)
            {
                foreach (IPEndPoint ipe in ipes)
                {
                    r = net.citrusleaf.Info.Get(ipe, names, out values, log);
                    if (r == CitrusleafResult.OK) return (r);
                }
            }
            return (r);
        }

        //----------------------------------------------------------------------------------------
        // Scan functionality.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Original scan callback delegate method.  This method will be called for each record
        /// returned from the scan.
        /// </summary>
        /// <param name="digest">Unique ID generated from the key and set name.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="generation">Generation indicates how many times the record has been modified.</param>
        /// <param name="valueCallbackPrivate">
        /// Pass-through user defined data. This field is not necessary because users can define non-static
        /// callback methods which contain class instance state. This method exists for backwards compatibility.
        /// </param>
        [Obsolete("Use IScanCallback instead.")]
        public delegate void ScanCallback(byte[] digest, IDictionary<string, object> values, uint generation, object valueCallbackPrivate);

        /// <summary>
        /// Alternative scan callback delegate method.  Allow user to terminate the scan by returning false.
        /// </summary>
        /// <param name="digest">Unique ID generated from the key and set name.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="generation">Generation indicates how many times the record has been modified.</param>
        /// <returns>true=continue scan, false=terminate scan</returns>
        [Obsolete("Use IScanCallback instead.")]
        public delegate bool ScanCallback2(byte[] digest, IDictionary<string, object> values, uint generation);

        /// <summary>
        /// This interface is used to support either old or new scan callback signatures.
        /// </summary>
        public interface IScanCallback
        {
            /// <summary>
            /// This method will be called for each record returned from the scan.
            /// The user can terminate the scan by returning false.
            /// </summary>
            /// <param name="ns">Namespace of the returned record.</param>
            /// <param name="set">Set of the returned record.</param>
            /// <param name="digest">Unique ID generated from the key and set name.</param>
            /// <param name="values">Bin name/value pairs.</param>
            /// <param name="generation">Generation indicates how many times the record has been modified.</param>
            /// <param name="expirationDate">Date this record will expire, in seconds from Jan 01 2010 00:00:00 GMT</param>
            /// <returns></returns>
            bool ScanRecordCallback(string ns, string set, byte[] digest, IDictionary<string, object> values, uint generation, uint expirationDate);
        }

        class ScanCallbackOld : IScanCallback
        {
            private ScanCallback callback;
            private object userData;

            public ScanCallbackOld(ScanCallback callback, object userData)
            {
                this.callback = callback;
                this.userData = userData;
            }

            bool IScanCallback.ScanRecordCallback(string ns, string set, byte[] digest, IDictionary<string, object> values, uint generation, uint expirationDate)
            {
                callback(digest, values, generation, userData);
                return true;
            }
        }

        class ScanCallbackOld2 : IScanCallback
        {
            private ScanCallback2 callback;

            public ScanCallbackOld2(ScanCallback2 callback)
            {
                this.callback = callback;
            }

            bool IScanCallback.ScanRecordCallback(string ns, string set, byte[] digest, IDictionary<string, object> values, uint generation, uint expirationDate)
            {
                return callback(digest, values, generation);
            }
        }

        /// <summary>
        /// Scan the database using specified namespace and set filters.  
        /// The supplied user-defined method is called for each record returned. 
        /// This method will block until the scan is complete.  Callbacks are 
        /// made within the scope of this method call.
        /// This method exists for backwards compatibility.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Set name. Pass null to retrieve all records in the namespace.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds. Not currently used.</param>
        /// <param name="valueCallback">Original callback delegate.</param>
        /// <param name="valueCallbackPrivate">
        /// Pass-through user defined data. This field is not necessary because users can define non-static
        /// callback methods which contain class instance state. This method exists for backwards compatibility.
        /// </param>
        /// <returns>Result status.</returns>
        [Obsolete("Use Scan with IScanCallback instead.")]
        public CitrusleafResult Scan(string ns, string set, int timeoutMillis, ScanCallback valueCallback, object valueCallbackPrivate)
        {
            return Scan(ns, set, timeoutMillis, new ScanCallbackOld(valueCallback, valueCallbackPrivate));
        }

        /// <summary>
        /// Scan the database using specified namespace and set filters.  
        /// The supplied user-defined method is called for each record returned. 
        /// The user can terminate the scan by returning false in the callback method.
        /// This method will block until the scan is complete or terminated by the callback
        /// method. Callbacks are made within the scope of this method call.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Set name. Pass null to retrieve all records in the namespace.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds. Not currently used.</param>
        /// <param name="callback">Callback delegate for each returned record.</param>
        /// <returns></returns>
        [Obsolete("Use Scan with IScanCallback instead.")]
        public CitrusleafResult Scan(string ns, string set, int timeoutMillis, ScanCallback2 callback)
        {
            return Scan(ns, set, timeoutMillis, new ScanCallbackOld2(callback));
        }

        /// <summary>
        /// Read records in the database using specified namespace filter, set name filter and the
        /// latest callback interface arguments.  This method issues scan requests to every server
        /// node in the cluster in parallel threads.  The client parses the returned scan data and
        /// calls the supplied user-defined callback method for each record returned.
        /// 
        /// The user can terminate the scan by returning false in the callback method.
        /// This method will block until the scan is complete or terminated by the callback
        /// method. Callbacks are made within the scope of this method call.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Set name. Pass null to retrieve all records in the namespace.</param>
        /// <param name="timeoutMillis">Timeout in milliseconds. Not currently used.</param>
        /// <param name="callback">Callback interface for each returned record.</param>
        /// <returns></returns>
        public CitrusleafResult Scan(string ns, string set, int timeoutMillis, IScanCallback callback)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already dispose");
            }

            if (asc == null)
                return CitrusleafResult.CLIENT_ERROR;

            Scan scan = new Scan(ns, set, callback, CreateValueFormatter(), log);
            return scan.ScanCluster(asc);
        }

        /// <summary>
        /// Scan a single node in the cluster.
        /// </summary>
        /// <param name="node">node can be obtained from GetActiveNodes().</param>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Set name. Pass null to retrieve all records in the namespace.</param>
        /// <param name="callback">Callback interface for each returned record.</param>
        /// <returns></returns>
        public CitrusleafResult ScanNode(Node node, string ns, string set, IScanCallback callback)
        {
            if (asc == null)
                return CitrusleafResult.CLIENT_ERROR;

            Scan scan = new Scan(ns, set, callback, CreateValueFormatter(), log);
            return scan.ScanNode(node);
        }

        //----------------------------------------------------------------------------------------
        // Database operations for RTA (real-time attribution) servers.
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Get bin value and generation for specified key.
        /// Requires a Citrusleaf RTA product backend.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="generation">Generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Transaction timeout in milliseconds.</param>
        /// <param name="max_results">Maximum number of records.</param>
        /// <param name="filters">Record filters.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetUsing(object key, out object value, out uint generation, int timeoutMillis, uint max_results, IList<ClAppendQueryFilter> filters)
        {
            return (GetUsing(DefaultNamespace, DefaultSet, key, DefaultBin, out value, out generation, timeoutMillis, max_results, filters));
        }

        /// <summary>
        /// Get bin value and generation for specified key and bin query filter.
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value</param>
        /// <param name="generation">Generation indicates how many times the record has been modified.</param>
        /// <param name="timeoutMillis">Client request timeout in milliseconds.</param>
        /// <param name="max_results">Maximum number of bins.</param>
        /// <param name="filters">Query filters.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult GetUsing
        (
            string ns, 
            string set, 
            object key, 
            string bin,
            out object value, 
            out uint generation,
            int timeoutMillis, 
            uint max_results, 
            IList<ClAppendQueryFilter> filters
        )
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // make sure there is a key
            value = null;
            generation = 0;

            if (key == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.Prepare(ns, set, key);
            b.AddAppendQuery(bin, max_results, filters);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            IDictionary<string, object> values;
            r = Do(partition, false, b, timeoutMillis, out values, out generation);

            if (r != CitrusleafResult.OK) return (r);
            if (values != null)
            {
                if (values.ContainsKey(bin))
                {
                    value = values[bin];
                }
                else
                {
                    return CitrusleafResult.BIN_NOT_FOUND_ERROR;
                }
            }

            return r;
        }

        /// <summary>
        /// Append value to record bin value already in the database for specified key. 
        /// Requires a Citrusleaf RTA product backend.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Append(object key, object value, CitrusleafWriteInfo writeInfo)
        {
            return (Append(DefaultNamespace, DefaultSet, key, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Append value to record bin value already in the database for specified key and bin name. 
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Append(string ns, string set, object key, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already dispose");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddAppend(bin, value);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            IDictionary<string, object> values;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Append each bin value in the dictionary to what's already in the database using 
        /// specified key and bin names. 
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Append(string ns, string set, object key, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddAppend(values);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Append each bin value in the list to what's already in the database for specified key
        /// and bin names.
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult Append(string ns, string set, object key, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // put two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (Append(ns, set, key, values_dict, writeInfo));
        }

        /// <summary>
        /// Append value to record bin value already in the database for specified key and bin name. 
        /// Supports additional append write options.
        /// Requires a Citrusleaf RTA product backend.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <param name="append_write_options">Append policy attributes.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AppendAs(object key, object value, CitrusleafWriteInfo writeInfo, ClAppendWriteOptions append_write_options)
        {
            return (AppendAs(DefaultNamespace, DefaultSet, key, DefaultBin, value, writeInfo, append_write_options));
        }

        /// <summary>
        /// Append value to record bin value already in the database for specified key and bin name. 
        /// Supports additional append write options.
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="set">Optional set name.</param>
        /// <param name="key">Record identifier. Unique within set.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <param name="append_write_options">Write policy attributes.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AppendAs(string ns, string set, object key, string bin, object value, CitrusleafWriteInfo writeInfo, ClAppendWriteOptions append_write_options)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.Prepare(ns, set, key);
            b.AddAppendExt(bin, value, append_write_options);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            IDictionary<string, object> values;

            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Append value to record bin value already in the database for specified key digest. 
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// Requires a Citrusleaf RTA product backend.
        /// The following properties must be set before using this method. 
        /// <list type="bullet">
        ///   <item><description>DefaultNamespace</description></item>
        ///   <item><description>DefaultSet</description></item>
        ///   <item><description>DefaultBin</description></item>
        /// </list> 
        /// </summary>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AppendDigest(byte[] keyDigest, object value, CitrusleafWriteInfo writeInfo)
        {
            return (AppendDigest(DefaultNamespace, keyDigest, DefaultBin, value, writeInfo));
        }

        /// <summary>
        /// Append value to record bin value already in the database for specified key digest and bin name. 
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bin">Bin name. Set to empty string if server nodes are configured as "single-bin".</param>
        /// <param name="value">Bin value.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AppendDigest(string ns, byte[] keyDigest, string bin, object value, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.AddAppend(bin, value);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            IDictionary<string, object> values;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        /// <summary>
        /// Append each bin value in the dictionary to what's already in the database using 
        /// specified key digest and bin names. 
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="values">Bin name/value pairs.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AppendDigest(string ns, byte[] keyDigest, IDictionary<string, object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.AddAppend(values);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));

        }

        /// <summary>
        /// Append each bin value in the list to what's already in the database using specified 
        /// key digest and bin names.
        /// Non-Digest database methods send the normal key and set name to the server which
        /// converts them to a digest. Digest methods send the digest directly.
        /// Requires a Citrusleaf RTA product backend.
        /// </summary>
        /// <param name="ns">Namespace.</param>
        /// <param name="keyDigest">Unique ID generated from the key and set name.</param>
        /// <param name="bins">Positionally matched bin names.</param>
        /// <param name="values">Positionally matched bin values.</param>
        /// <param name="writeInfo">Write policy attributes. Set to null to pick up cluster defaults.</param>
        /// <returns>Result status.</returns>
        public CitrusleafResult AppendDigest(string ns, byte[] keyDigest, IList<string> bins, IList<object> values, CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // shortcut for now - just zip the two together into a dict
            if (bins.Count != values.Count)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: bad parameters passed to set: bins and values list must have the same length");
                return (CitrusleafResult.CLIENT_ERROR);
            }
            IDictionary<string, object> values_dict = new Dictionary<string, object>();
            for (int i = 0; i < bins.Count; i++)
                values_dict[bins[i]] = values[i];
            return (AppendDigest(ns, keyDigest, values_dict, writeInfo));
        }

        /// <summary>
        /// Get list of active nodes in the cluster.
        /// </summary>
        public List<Node> GetActiveNodes()
        {
            return asc.GetActiveNodes();
        }

        //----------------------------------------------------------------------------------------
        // Internal methods.
        //----------------------------------------------------------------------------------------

        private CitrusleafResult SetInternal(string ns, string set, object key, IDictionary<string, object> values, bool uniq, bool binUniq, CitrusleafWriteInfo writeInfo) 
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // make sure there is a key
            if (key == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            // make sure there is a value
            if (values == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: value cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, set, key);
            b.AddWrite(values);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            if (uniq) b.set_Unique();
            if (binUniq) b.set_BinUnique();
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        private CitrusleafResult SetDigestInternal(string ns, byte[] keyDigest, IDictionary<string, object> values, bool uniq, bool binUniq,CitrusleafWriteInfo writeInfo)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("citrusleaf client already disposed");
            }
            // make sure there is a key
            if (keyDigest == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: key cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            // make sure there is a value
            if (values == null)
            {
                log.Log(LogLevel.ERROR, " Citrusleaf: value cannot be null");
                return CitrusleafResult.CLIENT_ERROR;
            }
            CitrusleafResult r;
            Buffer b = new Buffer(log);
            b.valueFormatter = CreateValueFormatter();
            b.Prepare(ns, keyDigest);
            b.AddWrite(values);
            if (writeInfo != null) b.set_WriteInfo(writeInfo);
            if (uniq) b.set_Unique();
            if (binUniq) b.set_BinUnique();
            uint partition;
            r = b.Ready(out partition);
            if (r != CitrusleafResult.OK) return (r);

            uint generation;
            return (Do(partition, true, b, (writeInfo != null) ? writeInfo.timeoutMillis : 0, out values, out generation));
        }

        private CitrusleafResult Do(uint partition, bool write, Buffer b, int timeoutMillis, out IDictionary<string, object> values, out uint generation)
        {
            values = null;
            generation = 0;

            if (asc == null)
                return CitrusleafResult.CLIENT_ERROR;

            // Use UtcNow to avoid daylight savings time switchover issues.
            DateTime start = DateTime.UtcNow;
            DateTime limit = start.AddMilliseconds(timeoutMillis);

            int iterations = 0;
            int failedSends = 0;
            int failedReceives = 0;
            int failedSockets = 0;
            int failedNodes = 0;

            do  
            {
                iterations++;

                Node node = asc.GetNode(b.ns, partition, write);
                if (node == null)
                {
                    failedNodes++;

                    // Fail if no timeout specified and GetNode will not succeed.
                    if (failedNodes >= 10)
                        break;

                    // Try 3 times before failing.
                    int sleepMillis = (timeoutMillis >= 3)? timeoutMillis / 3 : 1;
                    Thread.Sleep(sleepMillis);  
                    continue;
                }

                Socket socket = node.GetSocket();
                if (socket == null)
                {
                    failedSockets++;
                    continue; // means node probably dead, pick another
                }

                if (timeoutMillis > 0)
                {
                    // Set socket timeout to difference from now till max time limit.
                    int t = (int)limit.Subtract(DateTime.UtcNow).TotalMilliseconds;
                    if (t <= 0)
                        break; // Already timed out.

                    // The socket class does not allow send timeouts < 500ms, so this
                    // may not have the desired effect.
                    socket.SendTimeout = t;
                    socket.ReceiveTimeout = t;
                }
                else
                {
                    socket.SendTimeout = 0;
                    socket.ReceiveTimeout = 0;
                }

                CitrusleafResult r = b.Put(socket);

                if (r == CitrusleafResult.TIMEOUT)
                    break;

                if (r == CitrusleafResult.SERVER_ERROR)
                {
                    failedSends++;
                    node.DecreaseHealth(10);
                    socket.Close();
                    continue;
                }

                r = b.Get(socket, out values, out generation);

                if (r == CitrusleafResult.TIMEOUT)
                    break;

                if (r == CitrusleafResult.SERVER_ERROR)
                {
                    failedReceives++;
                    node.DecreaseHealth(10);
                    socket.Close();
                    continue;
                }
                node.PutSocket(socket);
                node.ResetHealth();
                return r;

            } while (DateTime.UtcNow <= limit);

            if (log.LevelEnabled(LogLevel.DEBUG))
            {
                log.Log(LogLevel.DEBUG, "Client timeout: timeout={0}ms iterations={1} failedSends={2} failedReceives={3} failedSockets={4} failedNodes={5}",
                    timeoutMillis, iterations, failedSends, failedReceives, failedSockets, failedNodes);
            }
            return CitrusleafResult.TIMEOUT;
        }

        //----------------------------------------------------------------------------------------
        // Logging
        //----------------------------------------------------------------------------------------

        /// <summary>
        /// Log escalation level.
        /// </summary>
        public enum LogLevel
        {
            /// <summary>
            /// Fatal error which prevents program from continuing.
            /// </summary>
            FATAL = 7,
            /// <summary>
            /// Severe error that needs immediate attention.
            /// </summary>
            CRITICAL = 6,
            /// <summary>
            /// Error conditon has occurred.
            /// </summary>
            ERROR = 5,
            /// <summary>
            /// Warning condition has occurred.
            /// </summary>
            WARNING = 4,
            /// <summary>
            /// Normal informational log message.
            /// </summary>
            INFO = 3,
            /// <summary>
            /// Message used for debugging purposes.
            /// </summary>
            DEBUG = 2,
            /// <summary>
            /// Descriptive message also used for debugging purposes.
            /// </summary>
            VERBOSE = 1
        };

        /// <summary>
        /// The log is written to the Windows Event Log by default.
        /// The caller can override this and set a private method which will be called 
        /// for each log entry. It is not legal to call Citrusleaf during the callback 
        /// because a deadlock could result.
        /// </summary>
        /// <param name="level">Log level.</param>
        /// <param name="msg">Log message.</param>
        /// <param name="uobject">User defined data passed with message.</param>
        public delegate void LogCallback(LogLevel level, string msg, object uobject);

        /// <summary>
        /// Set the logging level of citrusleaf. Allows the specification of a callback delegate,
        /// which will be called whenever a message of greater verbosity than the set level happens.
        /// </summary>
        /// <param name="level">Log level.</param>
        /// <param name="callback">Method to call.</param>
        /// <param name="uobject">User defined data passed with message.</param>
        public void SetLogging(LogLevel level, LogCallback callback, object uobject)
        {
            log.Set(level, callback, uobject);
        }

        /// <summary>
        /// Factory method for creating formatters.
        /// </summary>
        public Func<System.Runtime.Serialization.IFormatter> CreateValueFormatter { get; set; }
    }
}
