﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

namespace net.citrusleaf
{
    /// <summary>
    /// Citrusleaf client logger.  The log is written to the Windows Event Log by default.
    /// The caller can override this and set a private method which will be called 
    /// for each log entry.
    /// </summary>
    public class ClLog
    {
        private CitrusleafClient.LogLevel level;
        private CitrusleafClient.LogCallback callback;
        Object uobject;

        /// <summary>
        /// Construct logger.
        /// </summary>
        public ClLog()
        {
            level = CitrusleafClient.LogLevel.WARNING;
            callback = null;
            uobject = null;
        }

        /// <summary>
        /// Convert Citrusleaf log level to Windows log level.
        /// </summary>
        /// <param name="in_level">Citrusleaf log level.</param>
        /// <returns>Windows log level.</returns>
        private EventLogEntryType ConvertLevel(CitrusleafClient.LogLevel in_level)
        {
            if (in_level == CitrusleafClient.LogLevel.INFO) return (EventLogEntryType.Information);
            if (in_level == CitrusleafClient.LogLevel.WARNING
                || in_level == CitrusleafClient.LogLevel.ERROR) return(EventLogEntryType.Warning);
            if (in_level == CitrusleafClient.LogLevel.CRITICAL
                || in_level == CitrusleafClient.LogLevel.FATAL) return (EventLogEntryType.Error);
            return(EventLogEntryType.Information);
        }

        /// <summary>
        /// Set the client logging level. Allows the specification of a callback delegate,
        /// which will be called whenever a message of greater verbosity than the set level happens.
        /// </summary>
        /// <param name="level">Log level.</param>
        /// <param name="callback">Method that will receive log message.</param>
        /// <param name="uobject">User defined pass through data.</param>
        public void Set(CitrusleafClient.LogLevel level,
                        CitrusleafClient.LogCallback callback,
                        Object uobject)
        {
            this.level = level;
            this.callback = callback;
            this.uobject = uobject;
        }

        /// <summary>
        /// Determine if given log level will result in a log message or not.
        /// </summary>
        /// <param name="level">Log level to check.</param>
        /// <returns>true: log message is allowed. false: log message will be filtered out.</returns>
        public bool LevelEnabled(CitrusleafClient.LogLevel level)
        {
            return level >= this.level;
        }

        /// <summary>
        /// Log message to user defined callback or Windows event log.
        /// </summary>
        /// <param name="level">Log level.</param>
        /// <param name="fmt">Message format.</param>
        /// <param name="argsRest">Message arguments.</param>
        public void Log(CitrusleafClient.LogLevel level, string fmt, params object[] argsRest)
        {
            if (level >= this.level)
            {
                string o = String.Format(fmt, argsRest);
                if (callback != null)
                {
                    callback(level, o, uobject);
                }
                else
                {
                    EventLog.WriteEntry("Application", o, ConvertLevel(level), 1001);
                }
            }
        }
    }
}
