using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;
using System.Collections.Concurrent;

namespace net.citrusleaf
{
    public class Cluster : IDisposable
    {
        private const int CRAWLER_SLEEP_MS = 2000;
        private const int CRAWLER_FAST_MS = 50;
        public ClLog log;
        public Dictionary<string, string> address_mapping;
        private object cluster_lock;
        private List<Node> node_list; 
        private IDictionary<string, Node> nodes;
        private Random random;
        private Thread crawler_thread;
        private Settings settings;

        protected class WrappedCluster: IDisposable
        {
            public int refcount;
            public  Cluster cluster;
    
            public    WrappedCluster(Cluster cluster)
            {
                this.cluster = cluster;
            }

            public void Dispose()
            {
                // pattern for disposing http://msdn.microsoft.com/en-us/library/fs2xkftw%28VS.80%29.aspx
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            ~WrappedCluster() 
            {
                Dispose(false);
            }

            private bool disposed = false;

            protected virtual void Dispose(bool disposing)
            {
               // Check to see if Dispose has already been called.
               if(!this.disposed)
               {
                 // If disposing equals true, dispose all managed 
                 // and unmanaged resources.
                 if(disposing)
                 {
                    // Dispose managed resources.
                    cluster.Dispose();
                 }
              }
              disposed = true;         
           }
        }

        protected class Host
        {
            public string hostname;
            public int port;
        }
        protected List<Host> host_list;

        // Allow explicit dispose
        // This is set to true only when the last Dispose() is called, causing reference count on "this" to be down to zero
        private bool disposed = false;

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ~Cluster()
        {
            Dispose(false);
        }

        private static ConcurrentDictionary<String, WrappedCluster> clusterList = new ConcurrentDictionary<String, WrappedCluster>();
        private static object clusterlist_lock = new Object();

        /// <summary>
        /// Try to get an existing cluster.
        /// </summary>
        public static Cluster tryGetClusterInstance(string host, int port,ClLog log)
        {
            lock (clusterlist_lock)
            {
                string key = Cluster.makeClusterListKey(host,port);
                WrappedCluster aCluster=null;
                if (clusterList.TryGetValue(key, out aCluster) == false)
                {
                    log.Log(CitrusleafClient.LogLevel.VERBOSE, "getCluster: no cluster of {0} to reuse " , key);
                    return null;
                }
                else
                {
                    aCluster.refcount++;
                    log.Log(CitrusleafClient.LogLevel.VERBOSE, "getCluster: reusing existing cluster for {0}, refcount= {1}", key, aCluster.refcount);
                    return aCluster.cluster;
                }
            }       
        }

        public static Cluster getOrAdd(Cluster c, string host, int port, ClLog log) 
        {
            lock (clusterlist_lock)
            {
                string key = Cluster.makeClusterListKey(host,port);
                WrappedCluster newCluster = new WrappedCluster(c);
                WrappedCluster aCluster = clusterList.GetOrAdd(key, newCluster);
                aCluster.refcount++;
                if (aCluster == newCluster)
                {
                    log.Log(CitrusleafClient.LogLevel.VERBOSE, "getCluster: adding new cluster for {0}, refcount={1}", key, aCluster.refcount);
                }
                else
                {
                    log.Log(CitrusleafClient.LogLevel.VERBOSE, "getCluster: reuse cluster for {0}, refcount={1}", key, aCluster.refcount);
                }
                return aCluster.cluster;
            }
        }

        private static String makeClusterListKey(String host, int port)
        {
            return host + "_" + port;
        }

        public static void removeClusterInstance(Cluster cluster,ClLog log)
        {
            lock (clusterlist_lock)
            {
                WrappedCluster aCluster = null;
                string key = Cluster.makeClusterListKey(cluster.host_list[0].hostname, cluster.host_list[0].port);
                if (clusterList.TryGetValue(key, out aCluster)==false)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, "removeCluster: removing a cluster not in the dictionary {0} dict_size= {1} " , key, Cluster.clusterList.Count);
                    return ;
                }
                // decrement refcount
                WrappedCluster removedCluster = null;
                aCluster.refcount--;
                log.Log(CitrusleafClient.LogLevel.VERBOSE, "removeCluster: reference count for {0} is now {1}", key,aCluster.refcount);
                if (aCluster.refcount == 0)
                {
                    clusterList.TryRemove(key, out removedCluster);
                    if (removedCluster != aCluster)
                    {
                        log.Log(CitrusleafClient.LogLevel.WARNING, "removeCluster: cluster removed not the same as original {0} ", key);
                        return;
                    }
                    log.Log(CitrusleafClient.LogLevel.VERBOSE, "removeCluster: cluster remove from clusterlist");
                    aCluster.cluster.Dispose();
                    aCluster.cluster = null;
                }
            }

            return;
        }

        protected virtual void Dispose(bool disposing) 
        {
            if (!disposed)
            {
                disposed = true;

                // Free your own state (unmanaged objects).
                // this is an unmanaged resource
                if (crawler_thread != null && crawler_thread.IsAlive == true)
                {
                    crawler_thread.Abort();
                }

                if (disposing)
                {
                    foreach (var node in node_list)
                    {
                        node.Dispose();
                    }
                }
            }
        }

        /// <summary>
        /// Constructor - sets all the private vaules to default
        /// And processes a host entry
        /// </summary>
        public Cluster(Settings settings, ClLog log, Dictionary<string,string> address_mapping, string host, int port, int timeoutMillis)
        {
            this.settings = settings;
            this.address_mapping = address_mapping;
            this.log = log;
            nodes = new Dictionary<string, Node>(10);
            node_list = new List<Node>();
            random = new Random();
            cluster_lock = new Object();

            host_list = new List<Host>();

            write_replicas = new ConcurrentDictionary<string, Node[]>();

            // first index is partition, then the array of replicas
            // read_replicas = new ConcurrentDictionary<string, List<Node>[]>();

            // Add initial host.
            if (AddHost(host, port) != CitrusleafResult.OK)
            {
                throw new Exception(string.Format("getCluster: failed adding host {0} {1}", host, port));
            }

            waitTillStabilized(timeoutMillis);

            crawler_thread = new Thread(new ParameterizedThreadStart(Cluster.Crawler));
            // Start thread which will periodically get partition information from the cluster
            crawler_thread.Name = "crawler";
            crawler_thread.IsBackground = true;
            crawler_thread.Start(this);
        }

        /// <summary>
        /// Tend the cluster until it has stabilized and return control.
        /// This helps avoid initial database request timeout issues when
        /// a large number of threads are initiated at client startup.
        /// 
        /// If the cluster has not stabilized by the timeout, return
        /// control as well.  Do not return an error since future 
        /// database requests may still succeed.
        /// </summary>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        private void waitTillStabilized(int timeoutMillis)
        {
            DateTime limit = DateTime.Now.AddMilliseconds(timeoutMillis);
            int count = node_list.Count;

            do
            {
                Tend(this);

                // Check to see if cluster has changed since the last Tend().
                // If not, assume cluster has stabilized and return.
                if (count == node_list.Count)
                    return;

                Thread.Sleep(1);
                count = node_list.Count;
            } while (DateTime.Now < limit);
        }

        /// <summary>
        /// Check to see if this node is currently represented in the cluster.
        /// </summary>
        private bool AddressInCluster(IPEndPoint ipe)
        {
            lock (cluster_lock)
            {
                foreach (Node node in node_list)
                {
                    if (node.AddressInNode(ipe) == true) return (true);
                }
            }
            return (false);
        }

        public CitrusleafResult AddNode(IPEndPoint ipe, ClLog log)
        {
            log.Log(CitrusleafClient.LogLevel.VERBOSE,"consider adding node at ip {0}",ipe.ToString() );
            if (disposed)
            {
                throw new ObjectDisposedException("cluster already dispose");
            }
            string node_name;
            CitrusleafResult r = Info.Get(ipe, "node", out node_name, log);
            if (r != CitrusleafResult.OK) return (r);

            lock (cluster_lock)
            {
                if (!nodes.ContainsKey(node_name))
                {
                    log.Log(CitrusleafClient.LogLevel.DEBUG,"adding node {0}",ipe.ToString() );
                    Node node = new Node(node_name, ipe, this, settings);
                    InsertNode(node);
                }
            }

            return (CitrusleafResult.OK);
        }

        public CitrusleafResult AddHost(string hostname, int port)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("cluster already dispose");
            }
            // Try to Get at least one node
            IPEndPoint[] ipes = Resolve(hostname, port, log, this.address_mapping);
            if (ipes != null)
            {
                foreach (IPEndPoint ipe in ipes)
                {
                    CitrusleafResult r = AddNode(ipe, log);
                    if (r != CitrusleafResult.OK)
                    {
                        log.Log(CitrusleafClient.LogLevel.INFO, "fail AddNode() err {0}, host {1}, ipe {2}", r, hostname, ipe.ToString());
                    }
                    else
                    {
                        log.Log(CitrusleafClient.LogLevel.INFO, "AddNode() host {0}, ipe {1}", hostname, ipe.ToString());
                    }
                }
            }
            if (nodes.Count == 0)
            {
                log.Log(CitrusleafClient.LogLevel.INFO, "no nodes added for host {0}", hostname);
                return (CitrusleafResult.SERVER_ERROR);
            }

            // stash
            Host host = new Host();
            host.hostname = hostname;
            host.port = port;
            lock (cluster_lock)
                host_list.Add(host);

            return (CitrusleafResult.OK);
        }

        /// <summary>
        ///  Must be called with cluster lock held
        /// </summary>
        /// <param name="node"></param>
        private void InsertNode(Node node)
        {
            nodes[node.name] = node;
            node_list.Add(node);
        }

        /// <summary>
        ///  Must be called with cluster lock held
        /// </summary>
        /// <param name="node"></param>
        private void RemoveNode(Node node)
        {
            PartitionClearNode(node);
            nodes.Remove(node.name);
            node_list.Remove(node);
        }

        // index: namespace (so there's actually few of these)
        // array is indexed by partition
        ConcurrentDictionary<string, Node[]> write_replicas; 

        // first index is partition, then the array of replicas
        //ConcurrentDictionary<string, List<Node>[]> read_replicas;

        // number of partitions across the cluster
        // @TODO this should be dynamic 
        int n_partitions = 4096;

        // Track which nodes handle which partitions
        public void PartitionClearNode(Node node)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("cluster already dispose");
            }
            log.Log(CitrusleafClient.LogLevel.DEBUG, "partition clear: node {0}", node.name);

            foreach (Node[] ns_array in write_replicas.Values) {
                for (int i=0;i<ns_array.Length;i++) {
                    if (ns_array[i] == node)
                        ns_array[i] = null;
                }
            }

            /*
            foreach (List<Node>[] node_list_a in read_replicas.Values) {
                foreach( List<Node> node_list in node_list_a ) {
                    if (node_list != null)
                        node_list.Remove(node);
                }
            }
            */
        }

        public void PartitionSetNode(string ns, int partition, bool write, Node node)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("cluster already dispose");
            }

            if (write)
            {
                Node[] ns_array = null;
                write_replicas.TryGetValue(ns,out ns_array);
                if (ns_array == null)
                {
                    ns_array = new Node[n_partitions];
                    write_replicas[ns] = ns_array;
                }
                ns_array[partition] = node;
            }
            /*
            else
            {
                List<Node>[] ns_a_a = null;
                read_replicas.TryGetValue(ns, out ns_a_a);
                if (ns_a_a == null)
                {
                    ns_a_a = new List<Node>[n_partitions];
                    read_replicas[ns] = ns_a_a;
                }
                List<Node> node_a = ns_a_a[partition];
                if (node_a == null)
                {
                    node_a = new List<Node>();
                    ns_a_a[partition] = node_a;
                }
                node_a.Add(node);
            }
            */
        }

        /// <summary>
        /// Entry point of the tracker thread. One per cluster. 
        /// </summary>
        private static void Crawler(object asc_o)
        {
            Cluster asc = (Cluster)asc_o;
            ClLog log = asc.log;

            log.Log(CitrusleafClient.LogLevel.DEBUG, "Crawler started");
            try
            {
                int recheck_time = CRAWLER_FAST_MS;

                do
                {
                    Thread.Sleep(recheck_time);
                    bool status = Tend(asc);
                    recheck_time = (status)? CRAWLER_SLEEP_MS : CRAWLER_FAST_MS;
                } while (true);
            }
            catch (ThreadAbortException ) {
                // this is the normal termination path
                if (!asc.disposed)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, "Cluster MNA MNA crawler thread aborted");
                }
            }
            catch (Exception e)
            {
                log.Log(CitrusleafClient.LogLevel.WARNING,
                    "Citrusleaf Libary Crawler Exception {0} stack trace {1} ",e.ToString(), e.StackTrace);
            }
        }

        /// <summary>
        /// Check health of all nodes in the cluster.
        /// </summary>
        /// <param name="asc">Cluster.</param>
        /// <returns>true=node checks succeeded. false=at least one node check failed.</returns>
        public static bool Tend(Cluster asc)
        {
            ClLog log = asc.log;
            bool status = true;

            if ((asc.nodes != null) && (asc.nodes.Count > 0))
            {
                // Deleting nodes which are already dunned
                lock (asc.cluster_lock)
                {
                    List<Node> delete_me = new List<Node>(0);
                    foreach (Node node in asc.nodes.Values)
                    {
                        if (node.IsDunned()) delete_me.Add(node);
                    }
                    foreach (Node node in delete_me)
                    {
                        log.Log(CitrusleafClient.LogLevel.WARNING, " Node no longer reachable by client {0} {1} {2}", node.name, node.GetAddresses().Count, node.GetAddresses()[0]);
                        asc.RemoveNode(node);
                    }
                    delete_me = null;
                }

                // loop over all nodes, give them a health check by making a network request
                Dictionary<IPEndPoint, int> new_services = new Dictionary<IPEndPoint, int>();
                try
                {
                    foreach (Node node in asc.nodes.Values)
                    {
                        string services_str;
                        node.Update(out services_str); // here's the health check!
                        if (services_str != null && services_str.Length > 0)
                        {
                            log.Log(CitrusleafClient.LogLevel.DEBUG, "Node {0} checked passed ", node.name);
                            string[] services = services_str.Split(';');
                            foreach (string service in services)
                            {
                                string[] x = service.Split(':');
                                IPEndPoint[] service_ipes = Resolve(x[0], Convert.ToInt32(x[1]), log, asc.address_mapping);

                                if (service_ipes != null)
                                {
                                    foreach (IPEndPoint ipe in service_ipes)
                                    {
                                        if (false == asc.AddressInCluster(ipe))
                                            new_services[ipe] = 1;
                                    }
                                }
                            }
                        }
                        else if (asc.nodes.Count > 1)
                        {
                            log.Log(CitrusleafClient.LogLevel.INFO, "Node {0} check failed ", node.name);
                            status = false;
                        }
                    }
                    foreach (IPEndPoint ipe in new_services.Keys)
                    {
                        asc.AddNode(ipe, log);
                        log.Log(CitrusleafClient.LogLevel.INFO, "Node {0} added ", ipe.ToString());
                    }
                }
                catch (ThreadAbortException)
                {
                    // this is the normal termination path
                    if (! asc.disposed)
                    {
                        log.Log(CitrusleafClient.LogLevel.WARNING, "Cluster crawler thread aborted");
                    }
                    return status;
                }
                catch (Exception e)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING,
                        "Citrusleaf Libary Crawler Node Health check Exception {0} stack trace {1} ", e.ToString(), e.StackTrace);
                }
            }

            // if there are no nodes anymore, fall back to the initial configured list of hosts,
            // and try one of those.
            if ((asc.node_list == null) || (asc.node_list.Count <= 0))
            {
                if (asc.host_list == null || asc.host_list.Count <= 0)
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, "there isn't an initial host list");
                }
                else
                {
                    log.Log(CitrusleafClient.LogLevel.WARNING, " crawler sees no nodes anymore, trying adding nodes using initial host list {0} {1}",
                        asc.host_list[0], asc.host_list.Count);
                }
                foreach (Host host in asc.host_list)
                {
                    IPEndPoint[] ipes = Resolve(host.hostname, host.port, log, asc.address_mapping);

                    if (ipes != null)
                    {
                        foreach (IPEndPoint ipe in ipes)
                        {
                            CitrusleafResult r = asc.AddNode(ipe, log);
                            if (r == CitrusleafResult.OK)
                            {
                                log.Log(CitrusleafClient.LogLevel.WARNING, " successfully re-established cluster connection via " + ipe.ToString());
                                break;
                            }
                            else
                            {
                                log.Log(CitrusleafClient.LogLevel.WARNING, " failed re-established cluster connection via " + ipe.ToString() + " retry in 2 sec");
                            }
                        }
                    }
                    if ((asc.node_list == null) || (asc.node_list.Count <= 0)) break;
                }
            }
            return status;
        }

        // randomly select a node
        private int node_offset_counter = 0;
        public Node GetNodeRandom()
        {
            if (disposed)
            {
                throw new ObjectDisposedException("cluster already dispose");
            }
            Node r;

            lock (cluster_lock)
            {
                int i = 0; 
                do
                {
                    if ((node_list == null) || (node_list.Count <= 0))
                    {
                        log.Log(CitrusleafClient.LogLevel.INFO,"no node list in cluster");
                        return (null);
                    }
                    if (i == node_list.Count)
                    {
                        log.Log(CitrusleafClient.LogLevel.INFO,"no active nodes in cluster");
                        return (null);
                    }
                    i++;
                    int o = node_offset_counter;
                    node_offset_counter++;
                    o = o % node_list.Count;
                    r = node_list[o];
                } while (r.IsDunned() == true);
            }

            return (r);
        }

        public List<Node> GetActiveNodes()
        {
            lock (cluster_lock)
            {
                List<Node> activeNodes = new List<Node>(node_list.Count);

                foreach (Node node in node_list)
                {
                    if (!node.IsDunned())
                    {
                        activeNodes.Add(node);
                    }
                }
                return activeNodes;
            }
        }

        public Node GetNode(string ns, uint partition, bool write)
        {
            if (disposed)
            {
                throw new ObjectDisposedException("cluster already dispose");
            }

            // Use write replicas exclusively.
            Node node = null;
            Node[] node_a = null;
            write_replicas.TryGetValue(ns, out node_a);
            if (node_a != null)
                node = node_a[partition % n_partitions];

            if (node == null)
                return (GetNodeRandom());

            return (node);
            /*
            if (write)
            {
                Node[] node_a = null;
                write_replicas.TryGetValue(ns, out node_a);
                if (node_a != null)
                    node = node_a[partition % n_partitions];
            }
            else
            {
                List<Node>[] node_l_a = null;
                read_replicas.TryGetValue(ns, out node_l_a);
                if (node_l_a != null)
                {
                    List<Node> node_l = node_l_a[partition % n_partitions];
                    if (node_l != null && node_l.Count > 0)
                    {
                        node = node_l[random.Next(node_l.Count)];
                    }
                }
            }
            if (node == null)
                return (GetNodeRandom());

            return (node);
            */
        }

        public static IPEndPoint[] Resolve(string host, int port, ClLog log, Dictionary <string, string> address_mappings=null)
        {
            IPAddress[] addresses;
            try
            {
                if (address_mappings != null && address_mappings.ContainsKey(host))
                {
                    addresses = CitrusleafSocket.GetHostAddresses(address_mappings[host], 1000);
                }
                else
                {
                    addresses = CitrusleafSocket.GetHostAddresses(host, 1000);
                }
            }
            catch (Exception e)
            {
                log.Log(CitrusleafClient.LogLevel.ERROR, e.Message);
                return (null);
            }
            IPEndPoint[] endpoints = new IPEndPoint[addresses.Length];
            int i = 0;
            foreach (IPAddress address in addresses)
            {
                endpoints[i] = new IPEndPoint(address, port);
                i++;
            }
            return (endpoints);
        }
    }
}
