using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Threading;
using System.Collections.Concurrent;

namespace net.citrusleaf
{
    public class Node : IDisposable
    {
        public string name;
        private ConcurrentQueue<CitrusleafSocket> socketQueue;
        private IList<IPEndPoint> addresses;
        private Cluster asc; // cluster this belongs to, "back pointer"
        
        private double maxIdleSocketSeconds;
        private int maxSocketCount;
        
        private uint paxos_generation;

        private const int FULL_HEALTH = 250;
        private int health = FULL_HEALTH;

        public IList<IPEndPoint> GetAddresses()
        {
            return addresses;
        }

        public void DecreaseHealth(int h)
        {
            //bool iDunned = false;
            lock (this)
            {
                health -= h;
                asc.log.Log(CitrusleafClient.LogLevel.DEBUG, " node {0} decreased health by {1} now {2}", name, h, health);
                if (health <= 0)
                {
                    dunned = true;
                    //iDunned = true;
                    asc.log.Log(CitrusleafClient.LogLevel.INFO, " node {0} dunned by {1} now {2}", name, h, health);
                }
            }
            //if (iDunned)
            //    asc.log.Log(CitrusleafClient.LogLevel.WARNING, " node {0} dunned health {1}", name, health);
        }

        public void ResetHealth()
        {
            //asc.log.Log(CitrusleafClient.LogLevel.DEBUG, " node {0} reset health {1}", name, health);
            lock (this)
            {
                health = FULL_HEALTH;
                if (dunned == true) dunned = false;
            }
        }

        private bool dunned;
        public bool IsDunned() { return (dunned); }

        private bool disposed = false;

        public void Dispose()
        {
            // asc.log.Log(CitrusleafClient.LogLevel.VERBOSE, "Node dispose called");
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ~Node()
        {
            Dispose (false);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposed)
            {
                // Free your own state (unmanaged objects).
                if (disposing)
                {
                    // Free other state (managed objects).
                    if (socketQueue != null)
                    {
                        CitrusleafSocket s = null;

                        while (socketQueue.TryDequeue(out s))
                        {
                            s.Dispose();
                        }
                    }
                }

                // Set large fields to null.
                socketQueue = null;
                addresses = null;
                asc = null;
                disposed = true;
            }
        }

        public Socket GetSocket()
        {
            CitrusleafSocket cs = null;

            while (socketQueue.TryDequeue(out cs))
            {
                Socket socket = cs.s;

                if (! socket.Connected)
                {
                    socket.Close();
                    continue;
                }

                if ((DateTime.Now - cs.timestamp).TotalSeconds > maxIdleSocketSeconds)
                {
                    socket.Close();
                    continue;
                }
                return socket;
            }
            return createSocket();
        }

        private Socket createSocket()
        {
            foreach (IPEndPoint ipe in addresses)
            {
                Socket socket = new Socket(ipe.AddressFamily, SocketType.Stream, ProtocolType.Tcp);

                try
                {
                    socket.NoDelay = true;
                    socket.Connect(ipe);
                    return socket;
                }
                catch (SocketException ex)
                {
                    asc.log.Log(CitrusleafClient.LogLevel.WARNING, "Socket connect failed. Try next ipe: node {0} {1} {2} [{3}]", name, ipe.ToString(), ex.Message, ex.ErrorCode);
                    DecreaseHealth(20);
                }
            }
            return null;
        }

        public void PutSocket(Socket s)
        {
            if (s == null)
            {	
	            return;
            }

            // The socket queue can briefly have more entries than the maximum. 
            // This means an explicit mutex lock is not necessary.
            if (socketQueue.Count < maxSocketCount)
            {
                socketQueue.Enqueue(new CitrusleafSocket(s));
            }
            else
            {
                if (asc.log.LevelEnabled(CitrusleafClient.LogLevel.VERBOSE))
                {
                    asc.log.Log(CitrusleafClient.LogLevel.VERBOSE, "Socket count for node {0} exceeds soft limit {1}", name, maxSocketCount);
                }
                s.Close();
            }
        }

        /// <summary>
        /// Check to see if this address is currently in use by this node
        /// </summary>
        public bool AddressInNode(IPEndPoint ipe)
        {
            foreach (IPEndPoint node_ipe in addresses)
            {
                if (node_ipe.ToString() == ipe.ToString()) return (true);
            }
            return (false);
        }

        /// <summary>
        /// Ping this node and make sure everything's still OK.
        /// It will return the current services list, which can be followed by the cluster 
        /// that initiated the ping.
        /// </summary>
        public void Update(out string services)
        {
            services = null;
            string[] names = new string[] { "node", "services", "partition-generation" };

            IDictionary<string, string> values = null;
            CitrusleafResult r = CitrusleafResult.SERVER_ERROR;
            foreach (IPEndPoint ipe in addresses)
            {
                r = net.citrusleaf.Info.Get(ipe, names, out values, asc.log);
                if (r == CitrusleafResult.OK) break;
            }
            if (r != CitrusleafResult.OK)
            {
                asc.log.Log(CitrusleafClient.LogLevel.DEBUG, " node {0} update failed: info get: error {1}", this.name, r);
                DecreaseHealth(25);
                return;
            }
            // check if node is what we expect
            if (values["node"] != name)
            {
                asc.log.Log(CitrusleafClient.LogLevel.WARNING, "node name changed! dunning.");
                DecreaseHealth(FULL_HEALTH);
                return;
            }

            // successful request with same name, must be healthy
            ResetHealth();

            services = values["services"];

            string pg = null;
            values.TryGetValue("partition-generation", out pg);
            if (pg != null)
            {
                if (Convert.ToUInt32(pg) != paxos_generation)
                {
                    UpdateReplicaList();
                    paxos_generation = Convert.ToUInt32(pg);
                }
            }
        }

        private void UpdateReplicaList()
        {
            // Only track write replicas.
            //string[] names = new string[] { "replicas-read", "replicas-write" };
            string[] names = new string[] { "replicas-write" };

            IDictionary<string, string> values = null;
            CitrusleafResult r = CitrusleafResult.SERVER_ERROR;
            foreach (IPEndPoint ipe in addresses)
            {
                r = net.citrusleaf.Info.Get(ipe, names, out values, asc.log);
                if (r == CitrusleafResult.OK) break;
            }
            if (r != CitrusleafResult.OK)
            {
                asc.log.Log(CitrusleafClient.LogLevel.DEBUG, " node {0} update failed: info get: error {1}", this.name, r);
                DecreaseHealth(25);
                return;
            }

            // Do not clear node from partition map.  All partitions will eventually be updated after
            // all nodes are refreshed.
            //asc.PartitionClearNode(this);

            /* Do not track read replicas.
            string rv = values["replicas-read"];
            if (rv != null)
            {
                string[] rep_a = rv.Split(';');
                foreach (string rep in rep_a)
                {
                    string[] ns_part = rep.Split(':');
                    asc.PartitionSetNode(ns_part[0], Convert.ToInt32(ns_part[1]), false, this);
                }
            }
            */
            string rv = values["replicas-write"];
            if (rv != null)
            {
                string[] rep_a = rv.Split(';');
                foreach (string rep in rep_a)
                {
                    string[] ns_part = rep.Split(':');
                    asc.PartitionSetNode(ns_part[0], Convert.ToInt32(ns_part[1]), true, this);
                }
            }
        }

        public Node(string name, IPEndPoint address, Cluster asc, Settings settings)
        {
            dunned = false;
            this.name = name;
            addresses = new List<IPEndPoint>();
            addresses.Add(address);
            this.asc = asc;
            socketQueue = new ConcurrentQueue<CitrusleafSocket>();
            maxSocketCount = settings.MaxSocketCount;
            maxIdleSocketSeconds = (double)settings.MaxIdleSocketSeconds;
        }
    }
}
