﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Net.Sockets;
using System.Runtime.Serialization;

namespace net.citrusleaf
{

    public class Scan
    {
		private string ns;
        private string set;
        private CitrusleafClient.IScanCallback callback;
        private IFormatter formatter;
        private ClLog log;
        private ScanThread[] threads;
		private CitrusleafResult result;

        public Scan
        (
            string ns,
            string set,
            CitrusleafClient.IScanCallback callback,
            IFormatter formatter, 
            ClLog log
        )
        {
            this.ns = ns;
            this.set = set;
            this.callback = callback;
            this.formatter = formatter;
            this.log = log;
			this.result = CitrusleafResult.OK;
       }

        /// <summary>
        /// Send scan request to every node in the cluster in parallel.
        /// </summary>
        public CitrusleafResult ScanCluster(Cluster cluster)
        {
			//log.Log(CitrusleafClient.LogLevel.INFO, "Scan Begin: " + Thread.CurrentThread.ManagedThreadId);

            List<Node> activeNodes = cluster.GetActiveNodes();

            if (activeNodes.Count <= 0)
                return CitrusleafResult.CLIENT_ERROR;

            threads = new ScanThread[activeNodes.Count];

            for (int i = 0; i < activeNodes.Count; i++)
            {
                ScanThread thread = new ScanThread(this, activeNodes[i]);
                threads[i] = thread;
            }

			foreach (ScanThread thread in threads)
			{
				thread.Start();
			} 
			
			foreach (ScanThread thread in threads)
            {
                thread.Join();

                if (result == CitrusleafResult.OK && thread.Result != CitrusleafResult.OK)
                {
					result = thread.Result;
                }
            }
			//log.Log(CitrusleafClient.LogLevel.INFO, "Scan End: " + Thread.CurrentThread.ManagedThreadId + ',' + result);
			return result;
        }

        /// <summary>
        /// Send scan request to a single node.
        /// </summary>
        public CitrusleafResult ScanNode(Node node)
        {
            ScanThread thread = new ScanThread(this, node);
            thread.Run();
            return thread.Result;
        }

        /// <summary>
        /// Interrupt scan threads.
        /// </summary>
        public void Interrupt()
        {
			// Only need to interrupt threads once.
			lock (this)
			{
				if (result == CitrusleafResult.SCAN_TERMINATED)
				{
					return;
				}
				result = CitrusleafResult.SCAN_TERMINATED;
			}

			// Interrupt scan threads other than the current scan thread.
			int currentThreadId = Thread.CurrentThread.ManagedThreadId;

            foreach (ScanThread thread in threads)
            {
				if (thread.thread.ManagedThreadId != currentThreadId)
				{
					thread.Running = false;

					//log.Log(CitrusleafClient.LogLevel.INFO, "Attempt Close Socket: " + Thread.CurrentThread.ManagedThreadId + ',' + thread.thread.ManagedThreadId);
					thread.CloseSocket();
				}
            }
        }

        public string Namespace
        {
            get { return ns; }
        }

        public string Set
        {
            get { return set; }
        }

        public CitrusleafClient.IScanCallback Callback
        {
            get { return callback; }
        }

        public IFormatter Formatter
        {
            get { return formatter; }
        }

        public ClLog Log
        {
            get { return log; }
        }
    }

    public class ScanThread
    {
        private Scan parent;
        private Node node;
        public Thread thread;
        private Socket socket;
        private CitrusleafResult result;
        private volatile bool running = true;

        public ScanThread(Scan parent, Node node)
        {
            this.parent = parent;
            this.node = node;
			this.thread = new Thread(Run);
		}

        public void Start()
        {
            thread.Start();
        }

        public void Run()
        {
			//parent.Log.Log(CitrusleafClient.LogLevel.INFO, "Scan Node Begin: " + Thread.CurrentThread.ManagedThreadId + ',' + node.name);
			
			try
            {
                socket = node.GetSocket();

                if (socket != null)
                {
                    result = ScanNode();
                }
                else
                {
                    result = CitrusleafResult.SERVER_NOT_AVAILABLE;
                }
            }
            catch (Exception ex)
            {
				// Close socket so server can't send any more scan data to client.
				CloseSocket();
				
				if (running)
				{
					// Traditional exception has occurred.
					parent.Log.Log(CitrusleafClient.LogLevel.ERROR, ex.Message);
					result = CitrusleafResult.CLIENT_ERROR;
				}
				else
				{
					// Scan was terminated by parent thread resulting in an exception.
					result = CitrusleafResult.SCAN_TERMINATED;
				}
            }
			running = false;
			//parent.Log.Log(CitrusleafClient.LogLevel.INFO, "Scan Node End: " + Thread.CurrentThread.ManagedThreadId + ',' + node.name + ',' + result);
		}

        private CitrusleafResult ScanNode()
        {
            Buffer b = new Buffer(parent.Log);
            b.valueFormatter = parent.Formatter;
            b.PrepareScan(parent.Namespace, parent.Set);
            b.set_ReadIterate();

            uint partition;
            CitrusleafResult r = b.Ready(out partition);

            if (r != CitrusleafResult.OK)
            {
                CloseSocket();
                return r;
            }
            r = ScanNode(b);
            b.Reset();
            return r;
        }

        private CitrusleafResult ScanNode(Buffer b)
        {
			if (!running)
			{
				CloseSocket();
				return CitrusleafResult.SCAN_TERMINATED;
			}

			socket.SendTimeout = 0;
			CitrusleafResult r = b.Put(socket);

            if (r != CitrusleafResult.OK)
            {
                node.DecreaseHealth(10);
                CloseSocket();
                return r;
            }

            socket.ReceiveTimeout = 0;
			r = b.GetScan(this, socket, parent.Callback);

            if (r == CitrusleafResult.KEY_NOT_FOUND_ERROR)
            {
                r = CitrusleafResult.OK;
            }

            if (r == CitrusleafResult.OK)
            {
                node.PutSocket(socket);
                node.ResetHealth();
                return r;
            }
            CloseSocket();

            if (r == CitrusleafResult.SCAN_TERMINATED)
            {
                // User requested scan termination.
                // Terminate all scan threads.
				//parent.Log.Log(CitrusleafClient.LogLevel.INFO, "Scan Node Terminated: " + Thread.CurrentThread.ManagedThreadId + ',' + node.name);
				parent.Interrupt();
            }
            else if (r == CitrusleafResult.SERVER_ERROR)
            {
                node.DecreaseHealth(10);
            }
            return r;
        }

        public void CloseSocket()
        {
			lock (thread)
			{
				if (socket != null)
				{
					//parent.Log.Log(CitrusleafClient.LogLevel.INFO, "Close Socket: " + Thread.CurrentThread.ManagedThreadId + ',' + thread.ManagedThreadId);
					socket.Close();
					socket = null;
				}
			}
        }

        public void Join()
        {
            thread.Join();
        }

        public CitrusleafResult Result
        {
            get{return result;}
        }

        public bool Running
        {
            get {return running;}
			set {running = value;}
        }
    }
}
