﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace net.citrusleaf
{
    /// <summary>
    /// Code indicating how write transactions are handled.
    /// </summary>
    public enum CitrusleafWritePolicy
    {
        /// <summary>
        /// If transaction fails, retry until specified timeout is reached.
        /// </summary>
        RETRY,
        /// <summary>
        /// Fail immediately if transaction fails.
        /// </summary>
        ONCE,
        /// <summary>
        /// Guarantee that a transaction is applied exactly once.
        /// The server will keep state information about the transaction.
        /// </summary>
        ASSURED
    };

    /// <summary>
    /// This class is used to specify how write transactions are performed.  
    /// The specification includes:
    /// <list type="bullet">
    ///   <item><description>Timeouts</description></item>
    ///   <item><description>Per-transaction object expiration.</description></item>
    ///   <item><description>Optimistic locking policy through the generation count feature.</description></item>
    /// </list> 
    /// </summary>
    public class CitrusleafWriteInfo
    {
        /// <summary>
        /// Write policy.
        /// </summary>
        public CitrusleafWritePolicy write_policy;
        /// <summary>
        /// Total timeout of the transaction in milliseconds.
        /// </summary>
        public int timeoutMillis;
        /// <summary>
        /// Seconds that object remains active before expiring.
        /// </summary>
        public int object_expiration_seconds;

        /// <summary>
        /// Construct defaults with a specified timeout.
        /// </summary>
        /// <param name="timeoutMillis">Timeout in milliseconds.</param>
        public CitrusleafWriteInfo(int timeoutMillis)
        {
            write_policy = CitrusleafWritePolicy.RETRY; // if a connect cannot be established, retry within the timeout period
            Generation = 0;
            generation_is_set = false;
            this.timeoutMillis = timeoutMillis;
            object_expiration_seconds = 0; // if not set, the server will set a default value based on its configuration 
        }

        private bool generation_is_set;
        private uint generation;
        /// <summary>
        /// Expected record generation. Generation indicates how many times the record has been
        /// modified. The write will fail if the expected generation is set and it's not the
        /// generation on the server.  If set, the expected generation should be zero when creating
        /// a new record.
        /// </summary>
        public uint Generation
        {
            get { return (generation); }
            set
            {
                generation_is_set = true;
                generation = value;
            }
        }

        /// <summary>
        /// Indicates if generation is set.
        /// </summary>
        public bool GenerationSet()
        {
            return (generation_is_set);
        }

        /// <summary>
        /// Set object expiration in seconds. The written object is scheduled to expire after this 
        /// specified interval.
        /// </summary>
        /// <param name="seconds">Seconds before object expiration.</param>
        public void ExpirationSet(int seconds)
        {
            object_expiration_seconds = seconds;
        }

        /// <summary>
        /// Set object expiration using local time. The written object is scheduled to expire at 
        /// this specified local time.
        /// </summary>
        /// <param name="expiration">Local Time of expiration.</param>
        public void ExpirationSetLocal(DateTime expiration)
        {
            object_expiration_seconds = (int)((DateTime.Now.ToFileTime() - expiration.ToFileTime()) * 10000000);
        }

        /// <summary>
        /// Set object expiration using UTC (Coordinated Universal Time) time. The written object 
        /// is scheduled to expire at this specified UTC time.
        /// </summary>
        /// <param name="expiration">UTC Time of expiration.</param>
        public void ExpirationSetUtc(DateTime expiration)
        {
            object_expiration_seconds = (int)((DateTime.UtcNow.ToFileTime() - expiration.ToFileTime()) * 10000000);
        }
    }
}
